"""API of otx.algorithms.common.adapters.mmdeploy."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os
import time
from collections.abc import Mapping
from copy import deepcopy
from functools import partial
from subprocess import CalledProcessError
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import mmcv
import numpy as np
import torch
from mmcv.parallel import collate, scatter

from .utils.mmdeploy import (
    is_mmdeploy_enabled,
    mmdeploy_init_model_helper,
    update_deploy_cfg,
)
from .utils.onnx import prepare_onnx_for_openvino
from .utils.utils import numpy_2_list

# pylint: disable=too-many-locals


class NaiveExporter:
    """NaiveExporter for non-mmdeploy export."""

    @staticmethod
    def export2openvino(
        output_dir: str,
        model_builder: Callable,
        cfg: mmcv.Config,
        input_data: Dict[Any, Any],
        *,
        precision: str = "FP32",
        model_name: str = "model",
        input_names: Optional[List[str]] = None,
        output_names: Optional[List[str]] = None,
        opset_version: int = 11,
        dynamic_axes: Optional[Dict[Any, Any]] = None,
        mo_transforms: str = "",
    ):
        """Function for exporting to openvino."""
        input_data = scatter(collate([input_data], samples_per_gpu=1), [-1])[0]

        model = model_builder(cfg)
        model = model.cpu().eval()
        dynamic_axes = dynamic_axes if dynamic_axes else dict()

        onnx_path = NaiveExporter.torch2onnx(
            output_dir,
            model,
            input_data,
            model_name=model_name,
            input_names=input_names,
            output_names=output_names,
            opset_version=opset_version,
            dynamic_axes=dynamic_axes,
        )

        def get_normalize_cfg(cfg):
            def _get_normalize_cfg(cfg_):
                out = []
                if isinstance(cfg_, Mapping):
                    for key, value in cfg_.items():
                        if isinstance(value, (Mapping, list)):
                            out += _get_normalize_cfg(value)
                        if key == "type" and value == "Normalize":
                            return [cfg_]
                elif isinstance(cfg_, list):
                    for value in cfg_:
                        if isinstance(value, (Mapping, list)):
                            out += _get_normalize_cfg(value)
                return out

            cfg = _get_normalize_cfg(cfg)
            assert len(cfg) == 1
            return cfg[0]

        mo_args = {}

        normalize_cfg = get_normalize_cfg(cfg.data.test)
        if normalize_cfg.get("mean", None) is not None:
            mo_args["mean_values"] = normalize_cfg.get("mean")
        if normalize_cfg.get("std", None) is not None:
            mo_args["scale_values"] = normalize_cfg.get("std")
        if normalize_cfg.get("to_rgb", False):
            mo_args["reverse_input_channels"] = None

        if precision == "FP16":
            mo_args["compress_to_fp16"] = None
        if mo_transforms:
            mo_args["transform"] = mo_transforms

        NaiveExporter.onnx2openvino(
            output_dir,
            onnx_path,
            model_name=model_name,
            **mo_args,
        )

    @staticmethod
    def torch2onnx(
        output_dir: str,
        model: torch.nn.Module,
        input_data: Dict[Any, Any],
        *,
        model_name: str = "model",
        input_names: Optional[List[str]] = None,
        output_names: Optional[List[str]] = None,
        opset_version: int = 11,
        dynamic_axes: Optional[Dict[Any, Any]] = None,
        verbose: bool = False,
        **onnx_options,
    ) -> str:
        """Function for torch to onnx exporting."""

        img_metas = input_data.get("img_metas")
        numpy_2_list(img_metas)
        imgs = input_data.get("img")
        model.forward = partial(model.forward, img_metas=img_metas, return_loss=False)

        onnx_file_name = model_name + ".onnx"
        dynamic_axes = dynamic_axes if dynamic_axes else dict()
        torch.onnx.export(
            model,
            imgs,
            os.path.join(output_dir, onnx_file_name),
            verbose=verbose,
            export_params=True,
            input_names=input_names,
            output_names=output_names,
            dynamic_axes=dynamic_axes,
            opset_version=opset_version,
            operator_export_type=torch.onnx.OperatorExportTypes.ONNX,
            **onnx_options,
        )
        model.__dict__.pop("forward")
        return os.path.join(output_dir, onnx_file_name)

    @staticmethod
    def onnx2openvino(
        output_dir: str,
        onnx_path: str,
        *,
        model_name: str = "model",
        **openvino_options,
    ) -> Tuple[str, str]:
        """Function for onnx to openvino exporting."""
        from otx.algorithms.common.utils import mo_wrapper

        mo_args = {
            "input_model": onnx_path,
            "output_dir": output_dir,
            "model_name": model_name,
        }
        mo_args.update(openvino_options)

        ret, msg = mo_wrapper.generate_ir(output_dir, output_dir, silent=False, **mo_args)
        if ret != 0:
            raise ValueError(msg)
        return (
            os.path.join(output_dir, model_name + ".xml"),
            os.path.join(output_dir, model_name + ".bin"),
        )


if is_mmdeploy_enabled():
    import mmdeploy.apis.openvino as openvino_api
    from mmdeploy.apis import build_task_processor, extract_model, torch2onnx
    from mmdeploy.apis.openvino import get_input_info_from_cfg, get_mo_options_from_cfg
    from mmdeploy.core import reset_mark_function_count

    # from mmdeploy.core import FUNCTION_REWRITER
    from mmdeploy.utils import get_ir_config, get_partition_config

    class MMdeployExporter:
        """MMdeployExporter for mmdeploy exporting."""

        @staticmethod
        def export2openvino(
            output_dir: str,
            model_builder: Callable,
            cfg: mmcv.Config,
            deploy_cfg: mmcv.Config,
            *,
            model_name: str = "model",
        ):
            """Function for exporting to openvino."""

            task_processor = build_task_processor(cfg, deploy_cfg, "cpu")

            def helper(*args, **kwargs):
                return mmdeploy_init_model_helper(*args, **kwargs, model_builder=model_builder)

            task_processor.__class__.init_pytorch_model = helper

            input_data_cfg = deploy_cfg.pop(
                "input_data",
                {"shape": (128, 128, 3), "file_path": None},
            )

            if input_data_cfg.get("file_path"):
                import cv2

                input_data = cv2.imread(input_data_cfg.get("file_path"))
                # image assumed to be RGB format under OTX
                input_data = cv2.cvtColor(input_data, cv2.COLOR_BGR2RGB)
            else:
                input_data = np.zeros(input_data_cfg["shape"], dtype=np.uint8)

            partition_cfgs = get_partition_config(deploy_cfg)
            if partition_cfgs:
                MMdeployExporter.extract_partition(
                    output_dir,
                    input_data,
                    cfg,
                    deploy_cfg,
                    model_name=model_name,
                )

            onnx_paths = []
            onnx_paths.append(
                MMdeployExporter.torch2onnx(
                    output_dir,
                    input_data,
                    cfg,
                    deploy_cfg,
                    model_name=model_name,
                )
            )

            for onnx_path in onnx_paths:
                deploy_cfg_ = deepcopy(deploy_cfg)
                update_deploy_cfg(onnx_path, deploy_cfg_)
                MMdeployExporter.onnx2openvino(
                    output_dir,
                    onnx_path,
                    deploy_cfg_,
                )

        @staticmethod
        def extract_partition(
            output_dir: str,
            input_data: Any,
            cfg: mmcv.Config,
            deploy_cfg: mmcv.Config,
            *,
            model_name: str = "model",
        ):
            """Function for extracting partition."""

            model_onnx = MMdeployExporter.torch2onnx(
                output_dir,
                input_data,
                cfg,
                deploy_cfg,
                model_name=model_name,
            )

            partition_cfgs = get_partition_config(deploy_cfg)
            partition_cfgs = partition_cfgs.get("partition_cfg", None)
            partition_onnx = MMdeployExporter.partition_onnx(
                output_dir,
                model_onnx,
                partition_cfgs,
            )

            deploy_cfg_ = deepcopy(deploy_cfg)
            update_deploy_cfg(partition_onnx[0], deploy_cfg_)
            MMdeployExporter.onnx2openvino(
                output_dir,
                partition_onnx[0],
                deploy_cfg_,
            )
            deploy_cfg["partition_config"]["apply_marks"] = False
            reset_mark_function_count()

        @staticmethod
        def torch2onnx(
            output_dir: str,
            input_data: Any,
            cfg: mmcv.Config,
            deploy_cfg: mmcv.Config,
            *,
            model_name: str = "model",
        ) -> str:
            """Function for torch to onnx exporting."""
            onnx_file_name = model_name + ".onnx"
            torch2onnx(
                input_data,
                output_dir,
                onnx_file_name,
                deploy_cfg=deploy_cfg,
                model_cfg=cfg,
                model_checkpoint=cfg.get("load_from", None),
                device="cpu",
            )
            return os.path.join(output_dir, onnx_file_name)

        @staticmethod
        def partition_onnx(
            output_dir,
            onnx_path: str,
            partition_cfgs: Union[mmcv.ConfigDict, List[mmcv.ConfigDict]],
        ) -> Tuple[str, ...]:
            """Function for parition onnx."""
            partitioned_paths = []

            if not isinstance(partition_cfgs, list):
                partition_cfgs = [partition_cfgs]

            for partition_cfg in partition_cfgs:
                save_file = partition_cfg["save_file"]
                save_path = os.path.join(output_dir, save_file)
                start = partition_cfg["start"]
                end = partition_cfg["end"]
                dynamic_axes = partition_cfg.get("dynamic_axes", None)

                extract_model(onnx_path, start, end, dynamic_axes=dynamic_axes, save_file=save_path)
                partitioned_paths.append(save_path)
            return tuple(partitioned_paths)

        @staticmethod
        def onnx2openvino(
            output_dir: str,
            onnx_path: str,
            deploy_cfg: Union[str, mmcv.Config],
            *,
            model_name: Optional[str] = None,
        ) -> Tuple[str, str]:
            """Function for onnx to openvino exporting."""

            input_info = get_input_info_from_cfg(deploy_cfg)
            output_names = get_ir_config(deploy_cfg).output_names
            mo_options = get_mo_options_from_cfg(deploy_cfg)

            if not model_name:
                model_name = os.path.basename(onnx_path).replace(".onnx", "")
            mo_options.args += f'--model_name "{model_name}" '

            onnx_ready_path = os.path.join(os.path.dirname(onnx_path), f"{model_name}_ready.onnx")
            prepare_onnx_for_openvino(onnx_path, os.path.join(os.path.dirname(onnx_path), f"{model_name}_ready.onnx"))

            try:
                openvino_api.from_onnx(onnx_ready_path, output_dir, input_info, output_names, mo_options)
            except CalledProcessError as e:
                # NOTE: mo returns non zero return code (245) even though it successfully generate IR
                cur_time = time.time()
                time_threshold = 5
                if not (
                    e.returncode == 245
                    and not {model_name + ".bin", model_name + ".xml"} - set(os.listdir(output_dir))
                    and (
                        os.path.getmtime(os.path.join(output_dir, model_name + ".bin")) - cur_time < time_threshold
                        and os.path.getmtime(os.path.join(output_dir, model_name + ".xml")) - cur_time < time_threshold
                    )
                ):
                    raise e

            return (
                os.path.join(output_dir, model_name + ".xml"),
                os.path.join(output_dir, model_name + ".bin"),
            )



import torch
from mmdet.models.builder import build_detector

from otx.algorithms.detection.adapters.mmdet.models.detectors.custom_atss_detector import (
    CustomATSS,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestCustomATSS:
    @e2e_pytest_unit
    def test_custom_atss_build(self):
        model_cfg = dict(
            type="CustomATSS",
            backbone=dict(
                avg_down=False,
                base_channels=64,
                conv_cfg=None,
                dcn=None,
                deep_stem=False,
                depth=18,
                dilations=(1, 1, 1, 1),
                frozen_stages=-1,
                in_channels=3,
                init_cfg=None,
                norm_cfg=dict(requires_grad=True, type="BN"),
                norm_eval=True,
                num_stages=4,
                out_indices=(0, 1, 2, 3),
                plugins=None,
                pretrained=None,
                stage_with_dcn=(False, False, False, False),
                stem_channels=None,
                strides=(1, 2, 2, 2),
                style="pytorch",
                type="mmdet.ResNet",
                with_cp=False,
                zero_init_residual=True,
            ),
            neck=dict(
                type="FPN",
                in_channels=[64, 128, 256, 512],
                out_channels=64,
                start_level=1,
                add_extra_convs="on_output",
                num_outs=5,
                relu_before_extra_convs=True,
            ),
            bbox_head=dict(
                type="CustomATSSHead",
                num_classes=2,
                in_channels=64,
                stacked_convs=4,
                feat_channels=64,
                anchor_generator=dict(
                    type="AnchorGenerator",
                    ratios=[1.0],
                    octave_base_scale=8,
                    scales_per_octave=1,
                    strides=[8, 16, 32, 64, 128],
                ),
                bbox_coder=dict(
                    type="DeltaXYWHBBoxCoder", target_means=[0.0, 0.0, 0.0, 0.0], target_stds=[0.1, 0.1, 0.2, 0.2]
                ),
                loss_cls=dict(type="FocalLoss", use_sigmoid=True, gamma=2.0, alpha=0.25, loss_weight=1.0),
                loss_bbox=dict(type="GIoULoss", loss_weight=2.0),
                loss_centerness=dict(type="CrossEntropyLoss", use_sigmoid=True, loss_weight=1.0),
                use_qfl=False,
                qfl_cfg=dict(type="QualityFocalLoss", use_sigmoid=True, beta=2.0, loss_weight=1.0),
            ),
        )

        model = build_detector(model_cfg)
        assert isinstance(model, CustomATSS)

    @e2e_pytest_unit
    def test_custom_atss_load_state_dict_pre_hook(self):
        chkpt_classes = ["person", "car"]
        model_classes = ["tree", "car", "person"]
        chkpt_dict = {
            "bbox_head.atss_cls.weight": torch.tensor(
                [
                    [1, 1, 1, 1],
                    [2, 2, 2, 2],
                ]
            ),
            "bbox_head.atss_cls.bias": torch.tensor(
                [
                    [1],
                    [2],
                ]
            ),
        }
        model_dict = {
            "bbox_head.atss_cls.weight": torch.tensor(
                [
                    [3, 3, 3, 3],
                    [4, 4, 4, 4],
                    [5, 5, 5, 5],
                ]
            ),
            "bbox_head.atss_cls.bias": torch.tensor(
                [
                    [3],
                    [4],
                    [5],
                ]
            ),
        }
        gt_dict = {
            "bbox_head.atss_cls.weight": torch.tensor(
                [
                    [3, 3, 3, 3],
                    [2, 2, 2, 2],
                    [1, 1, 1, 1],
                ]
            ),
            "bbox_head.atss_cls.bias": torch.tensor(
                [
                    [3],
                    [2],
                    [1],
                ]
            ),
        }

        class Model:
            def state_dict(self):
                return model_dict

        CustomATSS.load_state_dict_pre_hook(Model(), model_classes, chkpt_classes, chkpt_dict, "")
        for k, gt in gt_dict.items():
            assert (chkpt_dict[k] != gt).sum() == 0



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-Lggggcense-Identifier: Apache-2.0
#

import pytest
import torch
from torch.nn import functional as F

from otx.core.ov.ops.normalizations import (
    MVNV6,
    BatchNormalizationV0,
    LocalResponseNormalizationV0,
    NormalizeL2V0,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestBatchNormalizationV0:
    @e2e_pytest_unit
    def test_forward(self):
        op = BatchNormalizationV0("dummy", shape=(1,), epsilon=1e-05, max_init_iter=2)
        op.eval()
        input = torch.randn(8, 128, 256, 256)
        gamma = torch.randn(128)
        beta = torch.randn(128)
        mean = torch.zeros(128)
        variance = torch.ones(128)

        output = op(input, gamma, beta, mean, variance)
        ref = F.batch_norm(input, mean, variance, gamma, beta, eps=op.attrs.epsilon)
        assert torch.equal(output, ref)

        op.train()
        outputs = []
        for _ in range(op.attrs.max_init_iter + 1):
            outputs.append(op(input, gamma, beta, mean, variance))
        assert torch.equal(outputs[0], outputs[1])
        assert not torch.equal(outputs[1], outputs[2])


class TestLocalResponseNormalizationV0:
    @e2e_pytest_unit
    def test_forward(self):
        op = LocalResponseNormalizationV0(
            "dummy",
            shape=(1,),
            alpha=0.0001,
            beta=0.75,
            bias=1.0,
            size=2,
        )
        input = torch.randn(6, 12, 10, 24)
        output = op(input, torch.tensor([1]))
        ref = F.local_response_norm(input, size=2, alpha=0.0001, beta=0.75, k=1.0)
        assert torch.equal(output, ref)


class TestNormalizeL2V0:
    @e2e_pytest_unit
    def test_invalid_attr(self):
        with pytest.raises(ValueError):
            NormalizeL2V0("dummy", shape=(1,), eps=0.01, eps_mode="error")

    @e2e_pytest_unit
    def test_forward(self):
        op = NormalizeL2V0("dummy", shape=(1,), eps=0.01, eps_mode="add")
        input = torch.rand(6, 12, 10, 24)

        output = op(input, torch.tensor([]))
        ref = input / (input + op.attrs.eps).sqrt()
        assert torch.equal(output, ref)

        output = op(input, torch.tensor([1]))
        ref = input / (input.pow(2).sum(1, keepdim=True) + op.attrs.eps).sqrt()
        assert torch.equal(output, ref)

        output = op(input, torch.tensor([1, 2, 3]))
        ref = input / (input.pow(2).sum([1, 2, 3], keepdim=True) + op.attrs.eps).sqrt()
        assert torch.equal(output, ref)

        output = op(input, torch.tensor([-2, -3]))
        ref = input / (input.pow(2).sum([-2, -3], keepdim=True) + op.attrs.eps).sqrt()
        assert torch.equal(output, ref)

        op = NormalizeL2V0("dummy", shape=(1,), eps=0.01, eps_mode="max")
        input = torch.rand(6, 12, 10, 24)

        output = op(input, torch.tensor([]))
        ref = input / torch.clamp(input, max=op.attrs.eps).sqrt()
        assert torch.equal(output, ref)

        output = op(input, torch.tensor([1]))
        ref = input / torch.clamp(input.pow(2).sum(1, keepdim=True), max=op.attrs.eps).sqrt()
        assert torch.equal(output, ref)

        output = op(input, torch.tensor([1, 2, 3]))
        ref = input / torch.clamp(input.pow(2).sum([1, 2, 3], keepdim=True), max=op.attrs.eps).sqrt()
        assert torch.equal(output, ref)

        output = op(input, torch.tensor([-2, -3]))
        ref = input / torch.clamp(input.pow(2).sum([-2, -3], keepdim=True), max=op.attrs.eps).sqrt()
        assert torch.equal(output, ref)


class TestMVNV6:
    @e2e_pytest_unit
    def test_invalid_attr(self):
        with pytest.raises(ValueError):
            MVNV6("dummy", shape=(1,), normalize_variance=True, eps=0.01, eps_mode="error")

    @e2e_pytest_unit
    def test_forward(self):
        op = MVNV6(
            "dummy",
            shape=(1,),
            normalize_variance=False,
            eps=0.01,
            eps_mode="INSIDE_SQRT",
        )
        input = torch.randn(6, 12, 10, 24)
        output = op(input, torch.tensor([1, 2]))
        ref = input - input.mean([1, 2], keepdim=True)
        assert torch.equal(output, ref)

        op = MVNV6(
            "dummy",
            shape=(1,),
            normalize_variance=True,
            eps=0.01,
            eps_mode="INSIDE_SQRT",
        )
        input = torch.randn(6, 12, 10, 24)
        output = op(input, torch.tensor([1, 2]))
        ref = input - input.mean([1, 2], keepdim=True)
        ref = ref / torch.sqrt(torch.square(ref).mean([1, 2], keepdim=True) + 0.01)
        assert torch.equal(output, ref)

        op = MVNV6(
            "dummy",
            shape=(1,),
            normalize_variance=True,
            eps=0.01,
            eps_mode="OUTSIDE_SQRT",
        )
        input = torch.randn(6, 12, 10, 24)
        output = op(input, torch.tensor([1, 2]))
        ref = input - input.mean([1, 2], keepdim=True)
        ref = ref / (torch.sqrt(torch.square(ref).mean([1, 2], keepdim=True)) + 0.01)
        assert torch.equal(output, ref)



"""Module for defining a semi-supervised multi-label classifier using mmcls."""
# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from mmcls.models.builder import CLASSIFIERS

from otx.algorithms.common.utils.logger import get_logger

from .sam_classifier import SAMImageClassifier

logger = get_logger()


@CLASSIFIERS.register_module()
class SemiSLMultilabelClassifier(SAMImageClassifier):
    """Semi-SL Multilabel Classifier which supports unlabeled data by overriding forward_train."""

    def forward_train(self, img, gt_label, **kwargs):
        """Data is transmitted as a classifier training function.

        Args:
            img (list[Tensor]): List of tensors of shape (1, C, H, W)
                Typically these should be mean centered and std scaled.
            gt_label (Tensor): Ground truth labels for the input labeled images
            kwargs (keyword arguments): Specific to concrete implementation
        """
        if "extra_0" not in kwargs:
            raise ValueError("'extra_0' does not exist in the dataset")
        if "img_strong" not in kwargs:
            raise ValueError("'img_strong' does not exist in the dataset")

        target = gt_label.squeeze()
        unlabeled_data = kwargs["extra_0"]
        x = {}
        x["labeled_weak"] = self.extract_feat(img)
        x["labeled_strong"] = self.extract_feat(kwargs["img_strong"])

        img_uw = unlabeled_data["img"]
        x["unlabeled_weak"] = self.extract_feat(img_uw)

        img_us = unlabeled_data["img_strong"]
        x["unlabeled_strong"] = self.extract_feat(img_us)

        losses = dict()
        loss = self.head.forward_train(x, target)
        losses.update(loss)

        return losses



"""Base supervised learning recipe for video classification."""

# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name

seed = 2

evaluation = dict(interval=1, metrics=["top_k_accuracy", "mean_class_accuracy"], final_metric="mean_class_accuracy")

optimizer = dict(
    type="AdamW",
    lr=0.001,
    weight_decay=0.0001,
)

optimizer_config = dict(grad_clip=dict(max_norm=40.0, norm_type=2))
lr_config = dict(
    policy="step",
    step=5,
    warmup="linear",
    warmup_by_epoch=True,
    warmup_iters=3,
)

# runtime settings
checkpoint_config = dict(interval=1)
log_config = dict(
    interval=10,
    hooks=[
        dict(type="TextLoggerHook", ignore_last=False),
    ],
)
# runtime settings
log_level = "INFO"
workflow = [("train", 1)]

find_unused_parameters = False
gpu_ids = range(0, 1)

dist_params = dict(backend="nccl")



"""Unit test for otx.algorithms.common.adapters.mmcv.hooks.early_stopping_hook."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from logging import Logger
from math import inf

import numpy as np
import pytest
from mmcv.runner import BaseRunner, LrUpdaterHook
from mmcv.utils import Config

from otx.algorithms.common.adapters.mmcv.hooks.early_stopping_hook import (
    EarlyStoppingHook,
    LazyEarlyStoppingHook,
    ReduceLROnPlateauLrUpdaterHook,
    StopLossNanTrainingHook,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class MockRunner(BaseRunner):
    """Mock class for BaseRunner."""

    def __init__(self) -> None:
        self._max_epochs = None
        self._iter = 9
        self._hooks = [LrUpdaterHook(warmup_iters=3)]
        self._rank = 0
        self.log_buffer = Config({"output": {"acc": 1.0}})
        self.logger = Logger("otx")
        self.should_stop = False
        self.optimizer = Config({"param_groups": [{"lr": 1e-4}]})
        self.bbox_mAP = 50.0

    def run(self):
        pass

    def save_checkpoint(self):
        pass

    def train(self):
        pass

    def val(self):
        pass


class TestEarlyStoppingHook:
    """Test class for EarlyStoppingHook."""

    @e2e_pytest_unit
    def test_init_rule(self) -> None:
        """Test funciton for init_rule function."""

        hook = EarlyStoppingHook(interval=5)
        with pytest.raises(KeyError):
            hook._init_rule("Invalid Key", "Invalid Indicator")
        with pytest.raises(ValueError):
            hook._init_rule(None, "Invalid Indicator")
        hook._init_rule("greater", "acc")
        assert hook.rule == "greater"
        assert hook.key_indicator == "acc"
        assert hook.compare_func(5, 9) is False
        hook._init_rule("less", "loss")
        assert hook.rule == "less"
        assert hook.key_indicator == "loss"
        assert hook.compare_func(5, 9) is True

    @e2e_pytest_unit
    def test_before_run(self) -> None:
        """Test function for before_run."""

        hook = EarlyStoppingHook(interval=5)
        runner = MockRunner()
        hook.before_run(runner)
        assert hook.by_epoch is False
        assert hook.warmup_iters == 3

        runner._hooks = []
        hook = EarlyStoppingHook(interval=5)
        with pytest.raises(ValueError):
            hook.before_run(runner)

    @e2e_pytest_unit
    def test_after_train_iter(self, mocker) -> None:
        """Test after_train_iter function."""

        mocker.patch.object(EarlyStoppingHook, "_do_check_stopping", return_value=True)
        hook = EarlyStoppingHook(interval=5)
        runner = MockRunner()
        hook.by_epoch = False
        hook.after_train_iter(runner)

    @e2e_pytest_unit
    def test_after_train_epoch(self, mocker) -> None:
        """Test after_train_epoch function."""

        mocker.patch.object(EarlyStoppingHook, "_do_check_stopping", return_value=True)
        runner = MockRunner()
        hook = EarlyStoppingHook(interval=5)
        hook.by_epoch = True
        hook.after_train_epoch(runner)

    @e2e_pytest_unit
    def test_do_check_stopping(self, mocker):
        """Test _do_check_stopping function."""

        runner = MockRunner()
        mocker.patch.object(EarlyStoppingHook, "_should_check_stopping", return_value=False)
        hook = EarlyStoppingHook(interval=5)
        hook.warmup_iters = 3
        hook._do_check_stopping(runner)
        mocker.patch.object(EarlyStoppingHook, "_should_check_stopping", return_value=True)
        runner.log_buffer = Config({"output": []})
        with pytest.raises(KeyError):
            hook._do_check_stopping(runner)
        runner = MockRunner()
        hook.key_indicator = "acc"
        hook._do_check_stopping(runner)
        assert hook.best_score == 1.0
        assert hook.wait_count == 0
        assert hook.last_iter == 9

        hook = EarlyStoppingHook(interval=5)
        hook.by_epoch = False
        hook.key_indicator = "acc"
        hook.warmup_iters = 3
        hook.best_score = 2.0
        hook.patience = 1.0
        hook.iteration_patience = 0.0
        hook._do_check_stopping(runner)
        assert hook.wait_count == 1
        assert runner.should_stop is True

        runner.should_stop = False
        hook = EarlyStoppingHook(interval=5)
        hook.by_epoch = False
        hook.key_indicator = "acc"
        hook.warmup_iters = 3
        hook.best_score = 2.0
        hook.patience = 1.0
        hook.iteration_patience = 20.0
        hook._do_check_stopping(runner)
        assert hook.wait_count == 1
        assert runner.should_stop is False

    @e2e_pytest_unit
    def test_should_check_stopping(self) -> None:
        """Test _should_check_stopping function."""

        hook = EarlyStoppingHook(interval=5)
        hook.by_epoch = False
        runner = MockRunner()
        assert hook._should_check_stopping(runner) is True

        runner._iter = 8
        assert hook._should_check_stopping(runner) is False


class TestLazyEarlyStoppingHook:
    """Test LazyEarlyStoppingHook function."""

    def test_should_check_stopping(self) -> None:
        hook = LazyEarlyStoppingHook(interval=5)
        hook.by_epoch = False

        runner = MockRunner()
        runner._iter = 8
        assert hook._should_check_stopping(runner) is False

        hook.start = 20
        assert hook._should_check_stopping(runner) is False

        runner._iter = 17
        hook.start = 4
        assert hook._should_check_stopping(runner) is False

        runner._iter = 18
        hook.start = 4
        assert hook._should_check_stopping(runner) is True


class TestReduceLROnPlateauLrUpdaterHook:
    """Test class for ReduceLROnPlateauLrUpdaterHook."""

    @e2e_pytest_unit
    def test_init_rule(self) -> None:
        """Test funciton for init_rule function."""

        hook = ReduceLROnPlateauLrUpdaterHook(interval=5, min_lr=1e-5)
        with pytest.raises(KeyError):
            hook._init_rule("Invalid Key", "Invalid Indicator")
        with pytest.raises(ValueError):
            hook._init_rule(None, "Invalid Indicator")
        hook._init_rule("greater", "acc")
        assert hook.rule == "greater"
        assert hook.key_indicator == "acc"
        assert hook.compare_func(5, 9) is False
        hook._init_rule("less", "loss")
        assert hook.rule == "less"
        assert hook.key_indicator == "loss"
        assert hook.compare_func(5, 9) is True

    @e2e_pytest_unit
    def test_is_check_timing(self) -> None:
        """Test _should_check_stopping function."""

        hook = ReduceLROnPlateauLrUpdaterHook(interval=5, min_lr=1e-5)
        hook.by_epoch = False
        runner = MockRunner()
        assert hook._is_check_timing(runner) is False

    @e2e_pytest_unit
    def test_get_lr(self, mocker) -> None:
        """Test function for get_lr."""

        mocker.patch.object(ReduceLROnPlateauLrUpdaterHook, "_is_check_timing", return_value=False)
        hook = ReduceLROnPlateauLrUpdaterHook(interval=5, min_lr=1e-5)
        hook.warmup_iters = 3
        runner = MockRunner()
        assert hook.get_lr(runner, 1e-2) == 1e-2

        mocker.patch.object(ReduceLROnPlateauLrUpdaterHook, "_is_check_timing", return_value=True)
        hook = ReduceLROnPlateauLrUpdaterHook(interval=5, min_lr=1e-5)
        hook.warmup_iters = 3
        runner = MockRunner()
        assert hook.get_lr(runner, 1e-2) == 1e-2
        assert hook.bad_count == 0

        mocker.patch.object(ReduceLROnPlateauLrUpdaterHook, "_is_check_timing", return_value=True)
        hook = ReduceLROnPlateauLrUpdaterHook(interval=5, min_lr=1e-5)
        hook.best_score = 90
        hook.warmup_iters = 3
        hook.bad_count = 2
        hook.iteration_patience = 5
        hook.last_iter = 8
        runner = MockRunner()
        assert hook.get_lr(runner, 1e-2) == 1e-2

        hook = ReduceLROnPlateauLrUpdaterHook(interval=5, min_lr=1e-5)
        hook.best_score = 90
        hook.warmup_iters = 3
        hook.bad_count = 2
        hook.iteration_patience = 5
        hook.last_iter = 2
        runner = MockRunner()
        assert hook.get_lr(runner, 1e-3) == 1e-3
        assert hook.last_iter == 2
        assert hook.bad_count == 2

    @e2e_pytest_unit
    def test_before_run(self) -> None:
        """Test function for before_run."""

        hook = ReduceLROnPlateauLrUpdaterHook(interval=5, min_lr=1e-5)
        runner = MockRunner()
        hook.before_run(runner)
        assert hook.base_lr == [1e-4]
        assert hook.bad_count == 0
        assert hook.last_iter == 0
        assert hook.current_lr == -1.0
        assert hook.best_score == -inf


class TestStopLossNanTrainingHook:
    """Test class for StopLossNanTrainingHook."""

    def test_after_train_iter(self) -> None:
        hook = StopLossNanTrainingHook()
        runner = MockRunner()
        runner.outputs = {"loss": np.array([np.nan])}
        hook.after_train_iter(runner)
        assert runner.should_stop is True



"""Test otx mmseg configurer."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import copy
import os

import pytest
import tempfile
from mmcv.utils import ConfigDict

from otx.algorithms.common.adapters.mmcv.utils.config_utils import MPAConfig
from otx.algorithms.segmentation.adapters.mmseg.configurer import (
    SegmentationConfigurer,
    IncrSegmentationConfigurer,
    SemiSLSegmentationConfigurer,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.segmentation.test_helpers import (
    DEFAULT_SEG_TEMPLATE_DIR,
)


class TestSegmentationConfigurer:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.configurer = SegmentationConfigurer()
        self.model_cfg = MPAConfig.fromfile(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "model.py"))
        self.data_cfg = MPAConfig.fromfile(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "data_pipeline.py"))

    @e2e_pytest_unit
    def test_configure(self, mocker):
        mock_cfg_base = mocker.patch.object(SegmentationConfigurer, "configure_base")
        mock_cfg_device = mocker.patch.object(SegmentationConfigurer, "configure_device")
        mock_cfg_ckpt = mocker.patch.object(SegmentationConfigurer, "configure_ckpt")
        mock_cfg_model = mocker.patch.object(SegmentationConfigurer, "configure_model")
        mock_cfg_task = mocker.patch.object(SegmentationConfigurer, "configure_task")
        mock_cfg_hook = mocker.patch.object(SegmentationConfigurer, "configure_hook")
        mock_cfg_gpu = mocker.patch.object(SegmentationConfigurer, "configure_samples_per_gpu")
        mock_cfg_fp16_optimizer = mocker.patch.object(SegmentationConfigurer, "configure_fp16_optimizer")
        mock_cfg_compat_cfg = mocker.patch.object(SegmentationConfigurer, "configure_compat_cfg")

        model_cfg = copy.deepcopy(self.model_cfg)
        data_cfg = copy.deepcopy(self.data_cfg)
        returned_value = self.configurer.configure(model_cfg, "", data_cfg, True)
        mock_cfg_base.assert_called_once_with(model_cfg, data_cfg, None, None)
        mock_cfg_device.assert_called_once_with(model_cfg, True)
        mock_cfg_model.assert_called_once_with(model_cfg, None)
        mock_cfg_ckpt.assert_called_once_with(model_cfg, "")
        mock_cfg_task.assert_called_once_with(model_cfg, True)
        mock_cfg_hook.assert_called_once_with(model_cfg)
        mock_cfg_gpu.assert_called_once_with(model_cfg, "train")
        mock_cfg_fp16_optimizer.assert_called_once_with(model_cfg)
        mock_cfg_compat_cfg.assert_called_once_with(model_cfg)
        assert returned_value == model_cfg

    @e2e_pytest_unit
    def test_configure_base(self, mocker):
        mocker.patch(
            "otx.algorithms.segmentation.adapters.mmseg.configurer.align_data_config_with_recipe",
            return_value=True,
        )

        model_cfg = copy.deepcopy(self.model_cfg)
        data_cfg = copy.deepcopy(self.data_cfg._cfg_dict)
        self.configurer.configure_base(model_cfg, data_cfg, [], [])

    @e2e_pytest_unit
    def test_configure_model(self):
        ir_options = {"ir_model_path": {"ir_weight_path": "", "ir_weight_init": ""}}
        self.configurer.configure_model(self.model_cfg, ir_options)
        assert self.model_cfg.model_task

    @e2e_pytest_unit
    def test_configure_data(self, mocker):
        data_cfg = copy.deepcopy(self.data_cfg)
        self.configurer.configure_data(self.model_cfg, True, data_cfg)
        assert self.model_cfg.data
        assert self.model_cfg.data.train
        assert self.model_cfg.data.val

    @e2e_pytest_unit
    def test_configure_task(self, mocker):
        model_cfg = copy.deepcopy(self.model_cfg)
        mock_cfg_classes = mocker.patch.object(SegmentationConfigurer, "configure_classes")
        mock_cfg_ignore = mocker.patch.object(SegmentationConfigurer, "configure_ignore")
        self.configurer.configure_task(model_cfg, True)

        mock_cfg_classes.assert_called_once()
        mock_cfg_ignore.assert_called_once()

    @e2e_pytest_unit
    def test_configure_ignore(self):
        model_cfg = copy.deepcopy(self.model_cfg)
        self.configurer.configure_ignore(model_cfg)
        if "decode_head" in model_cfg.model:
            assert model_cfg.model.decode_head.loss_decode.type == "CrossEntropyLossWithIgnore"

    @e2e_pytest_unit
    def test_configure_classes_replace(self, mocker):
        model_cfg = copy.deepcopy(self.model_cfg)
        self.configurer.task_adapt_op = "REPLACE"
        mocker.patch.object(SegmentationConfigurer, "get_data_classes", return_value=["foo", "bar"])
        self.configurer.configure_classes(model_cfg)
        assert "background" in self.configurer.model_classes
        assert self.configurer.model_classes == ["background", "foo", "bar"]

    @e2e_pytest_unit
    def test_configure_classes_merge(self, mocker):
        model_cfg = copy.deepcopy(self.model_cfg)
        self.configurer.task_adapt_op = "MERGE"
        mocker.patch.object(SegmentationConfigurer, "get_model_classes", return_value=["foo", "bar"])
        mocker.patch.object(SegmentationConfigurer, "get_data_classes", return_value=["foo", "baz"])
        self.configurer.configure_classes(model_cfg)
        assert "background" in self.configurer.model_classes
        assert self.configurer.model_classes == ["background", "foo", "bar", "baz"]

    @e2e_pytest_unit
    def test_configure_ckpt(self, mocker):
        model_cfg = copy.deepcopy(self.model_cfg)
        model_cfg.resume = True

        mocker.patch(
            "otx.algorithms.segmentation.adapters.mmseg.configurer.CheckpointLoader.load_checkpoint",
            return_value={"model": None},
        )
        with tempfile.TemporaryDirectory() as tempdir:
            self.configurer.configure_ckpt(model_cfg, os.path.join(tempdir, "dummy.pth"))

    @e2e_pytest_unit
    def test_configure_device(self, mocker):
        mocker.patch(
            "torch.distributed.is_initialized",
            return_value=True,
        )
        mocker.patch("os.environ", return_value={"LOCAL_RANK": 2})
        config = copy.deepcopy(self.model_cfg)
        self.configurer.configure_device(config, True)
        assert config.distributed is True

        mocker.patch(
            "torch.distributed.is_initialized",
            return_value=False,
        )
        mocker.patch(
            "torch.cuda.is_available",
            return_value=False,
        )
        config = copy.deepcopy(self.model_cfg)
        self.configurer.configure_device(config, True)
        assert config.distributed is False
        assert config.device == "cpu"

        mocker.patch(
            "torch.distributed.is_initialized",
            return_value=False,
        )
        mocker.patch(
            "torch.cuda.is_available",
            return_value=True,
        )
        config = copy.deepcopy(self.model_cfg)
        self.configurer.configure_device(config, True)
        assert config.distributed is False
        assert config.device == "cuda"

    @e2e_pytest_unit
    def test_configure_samples_per_gpu(self):
        model_cfg = copy.deepcopy(self.model_cfg)
        model_cfg.data.train.dataset.otx_dataset = range(1)
        self.configurer.configure_samples_per_gpu(model_cfg, "train")
        assert model_cfg.data.train_dataloader == {"samples_per_gpu": 1, "drop_last": True}

    @e2e_pytest_unit
    def test_configure_fp16_optimizer(self):
        model_cfg = copy.deepcopy(self.model_cfg)
        model_cfg.fp16 = {}
        self.configurer.configure_fp16_optimizer(model_cfg)
        assert model_cfg.optimizer_config.type == "Fp16OptimizerHook"

        model_cfg.fp16 = {}
        model_cfg.optimizer_config.type = "SAMOptimizerHook"
        self.configurer.configure_fp16_optimizer(model_cfg)
        assert model_cfg.optimizer_config.type == "Fp16SAMOptimizerHook"

        model_cfg.fp16 = {}
        model_cfg.optimizer_config.type = "DummyOptimizerHook"
        self.configurer.configure_fp16_optimizer(model_cfg)
        assert model_cfg.optimizer_config.type == "DummyOptimizerHook"

    @e2e_pytest_unit
    def test_configure_compat_cfg(self):
        model_cfg = copy.deepcopy(self.model_cfg)
        model_cfg.data.train_dataloader = {}
        model_cfg.data.val_dataloader = {}
        model_cfg.data.test_dataloader = {}
        self.configurer.configure_compat_cfg(model_cfg)

    @e2e_pytest_unit
    def test_configure_hook(self):
        model_cfg = copy.deepcopy(self.model_cfg)
        model_cfg.custom_hook_options = {"LazyEarlyStoppingHook": {"start": 5}, "LoggerReplaceHook": {"_delete_": True}}
        self.configurer.configure_hook(model_cfg)
        assert model_cfg.custom_hooks[0]["start"] == 5


class TestIncrSegmentationConfigurer:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.configurer = IncrSegmentationConfigurer()
        self.model_cfg = MPAConfig.fromfile(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "model.py"))
        self.data_cfg = MPAConfig.fromfile(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "data_pipeline.py"))

    @e2e_pytest_unit
    def test_configure_task(self, mocker):
        mocker.patch.object(SegmentationConfigurer, "configure_task")
        self.configurer.configure_task(self.model_cfg, True)
        assert self.model_cfg.custom_hooks[1].type == "TaskAdaptHook"
        assert self.model_cfg.custom_hooks[1].sampler_flag is False


class TestSemiSLSegmentationConfigurer:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.configurer = SemiSLSegmentationConfigurer()
        self.model_cfg = MPAConfig.fromfile(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "model.py"))
        self.data_cfg = MPAConfig.fromfile(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "data_pipeline.py"))

    @e2e_pytest_unit
    def test_configure_data(self, mocker):
        data_cfg = copy.deepcopy(self.data_cfg)
        unlabeled_dict = dict(data=dict(unlabeled=dict(otx_dataset="foo")))
        data_cfg.merge_from_dict(unlabeled_dict)
        mock_ul_dataloader = mocker.patch.object(SemiSLSegmentationConfigurer, "configure_unlabeled_dataloader")
        self.configurer.configure_data(self.model_cfg, True, data_cfg)
        mock_ul_dataloader.assert_called_once()

    @e2e_pytest_unit
    def test_configure_task(self, mocker):
        model_cfg = ConfigDict(dict(model=dict(type="", task_adapt=True)))
        mock_remove_hook = mocker.patch("otx.algorithms.segmentation.adapters.mmseg.configurer.remove_custom_hook")
        self.configurer.configure_task(model_cfg, True)
        mock_remove_hook.assert_called_once()

    @e2e_pytest_unit
    def test_configure_unlabeled_dataloader(self, mocker):
        data_cfg = copy.deepcopy(self.data_cfg)
        data_cfg.model_task = "segmentation"
        data_cfg.distributed = False
        unlabeled_dict = dict(data=dict(unlabeled=dict(otx_dataset="foo")))
        data_cfg.merge_from_dict(unlabeled_dict)
        mocker_build_dataset = mocker.patch("otx.algorithms.segmentation.adapters.mmseg.configurer.build_dataset")
        mocker_build_dataloader = mocker.patch("otx.algorithms.segmentation.adapters.mmseg.configurer.build_dataloader")
        self.configurer.configure_data(self.model_cfg, True, data_cfg)

        mocker_build_dataset.assert_called_once()
        mocker_build_dataloader.assert_called_once()
        assert "ComposedDataLoadersHook" in [hook["type"] for hook in self.model_cfg.custom_hooks]



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os

import numpy as np
import pytest
import torch
from mmcls.models import build_classifier

from otx.algorithms.classification.adapters.mmcls.configurer import ClassificationConfigurer
from otx.algorithms.common.adapters.mmcv.hooks.recording_forward_hook import (
    ReciproCAMHook,
)
from otx.algorithms.common.adapters.mmcv.utils.config_utils import MPAConfig
from otx.cli.registry import Registry
from tests.test_suite.e2e_test_system import e2e_pytest_unit

templates_cls = Registry("otx/algorithms").filter(task_type="CLASSIFICATION").templates
templates_cls_ids = [template.model_template_id for template in templates_cls]


class TestExplainMethods:
    ref_saliency_vals_cls = {
        "EfficientNet-B0": np.array([57, 0, 161, 127, 102, 96, 92], dtype=np.uint8),
        "MobileNet-V3-large-1x": np.array([140, 82, 87, 81, 79, 117, 254], dtype=np.uint8),
        "EfficientNet-V2-S": np.array([125, 42, 24, 21, 27, 55, 145], dtype=np.uint8),
    }

    @e2e_pytest_unit
    @pytest.mark.parametrize("template", templates_cls, ids=templates_cls_ids)
    def test_saliency_map_cls(self, template):
        torch.manual_seed(0)

        base_dir = os.path.abspath(os.path.dirname(template.model_template_path))
        cfg_path = os.path.join(base_dir, "model.py")
        cfg = MPAConfig.fromfile(cfg_path)

        cfg.model.pop("task")
        ClassificationConfigurer.configure_in_channel(cfg)
        model = build_classifier(cfg.model)
        model = model.eval()

        img = torch.ones(2, 3, 224, 224) - 0.5
        data = {"img_metas": {}, "img": img}

        with ReciproCAMHook(model) as rcam_hook:
            with torch.no_grad():
                _ = model(return_loss=False, **data)
        saliency_maps = rcam_hook.records

        assert len(saliency_maps) == 2
        assert saliency_maps[0].ndim == 3
        assert saliency_maps[0].shape == (1000, 7, 7)
        assert (saliency_maps[0][0][0] == self.ref_saliency_vals_cls[template.name]).all()



"""Memory cache handler implementations and singleton class to call them."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import ctypes as ct
import multiprocessing as mp
from multiprocessing.managers import DictProxy
from typing import Any, Dict, Optional, Union

import numpy as np
from mmcv.runner import get_dist_info
from multiprocess.synchronize import Lock

from otx.algorithms.common.utils.logger import get_logger

logger = get_logger()


class _DummyLock:
    def __enter__(self, *args, **kwargs):
        pass

    def __exit__(self, *args, **kwargs):
        pass


class MemCacheHandlerBase:
    """Base class for memory cache handler.

    It will be combined with LoadImageFromOTXDataset to store/retrieve the samples in memory.
    """

    def __init__(self, mem_size: int):
        self._init_data_structs(mem_size)

    def _init_data_structs(self, mem_size: int):
        self._arr = (ct.c_uint8 * mem_size)()
        self._cur_page = ct.c_size_t(0)
        self._cache_addr: Union[Dict, DictProxy] = {}
        self._lock: Union[Lock, _DummyLock] = _DummyLock()
        self._freeze = ct.c_bool(False)

    def __len__(self):
        """Get the number of cached items."""
        return len(self._cache_addr)

    @property
    def mem_size(self) -> int:
        """Get the reserved memory pool size (bytes)."""
        return len(self._arr)

    def get(self, key: Any) -> Optional[np.ndarray]:
        """Try to look up the cached item with the given key.

        Args:
            key (Any): A key for looking up the cached item

        Returns:
            If succeed return np.ndarray, otherwise return None
        """
        if self.mem_size == 0 or key not in self._cache_addr:
            return None

        addr = self._cache_addr[key]

        offset, count, shape, strides = addr

        data = np.frombuffer(self._arr, dtype=np.uint8, count=count, offset=offset)
        return np.lib.stride_tricks.as_strided(data, shape, strides)

    def put(self, key: Any, data: np.ndarray) -> Optional[int]:
        """Try to store np.ndarray with a key to the reserved memory pool.

        Args:
            key (Any): A key to store the cached item
            data (np.ndarray): A data sample to store

        Returns:
            Optional[int]: If succeed return the address of cached item in memory pool
        """
        if self._freeze.value:
            return None

        assert data.dtype == np.uint8

        with self._lock:
            new_page = self._cur_page.value + data.size

            if key in self._cache_addr or new_page > self.mem_size:
                return None

            offset = ct.byref(self._arr, self._cur_page.value)
            ct.memmove(offset, data.ctypes.data, data.size)

            self._cache_addr[key] = (
                self._cur_page.value,
                data.size,
                data.shape,
                data.strides,
            )
            self._cur_page.value = new_page
            return new_page

    def __repr__(self):
        """Representation for the current handler status."""
        perc = 100.0 * self._cur_page.value / self.mem_size if self.mem_size > 0 else 0.0
        return (
            f"{self.__class__.__name__} "
            f"uses {self._cur_page.value} / {self.mem_size} ({perc:.1f}%) memory pool and "
            f"store {len(self)} items."
        )

    def freeze(self):
        """If frozen, it is impossible to store a new item anymore."""
        self._freeze.value = True

    def unfreeze(self):
        """If unfrozen, it is possible to store a new item."""
        self._freeze.value = False


class MemCacheHandlerForSP(MemCacheHandlerBase):
    """Memory caching handler for single processing.

    Use if PyTorch's DataLoader.num_workers == 0.
    """


class MemCacheHandlerForMP(MemCacheHandlerBase):
    """Memory caching handler for multi processing.

    Use if PyTorch's DataLoader.num_workers > 0.
    """

    def _init_data_structs(self, mem_size: int):
        self._arr = mp.Array(ct.c_uint8, mem_size, lock=False)
        self._cur_page = mp.Value(ct.c_size_t, 0, lock=False)

        self._manager = mp.Manager()
        self._cache_addr: DictProxy = self._manager.dict()
        self._lock = mp.Lock()
        self._freeze = mp.Value(ct.c_bool, False, lock=False)

    def __del__(self):
        """When deleting, manager should also be shutdowned."""
        self._manager.shutdown()


class MemCacheHandlerError(Exception):
    """Exception class for MemCacheHandler."""


class MemCacheHandlerSingleton:
    """A singleton class to create, delete and get MemCacheHandlerBase."""

    instance: MemCacheHandlerBase

    @classmethod
    def get(cls) -> MemCacheHandlerBase:
        """Get the created MemCacheHandlerBase.

        If no one is created before, raise RuntimeError.
        """
        if not hasattr(cls, "instance"):
            cls_name = cls.__class__.__name__
            raise MemCacheHandlerError(f"Before calling {cls_name}.get(), you should call {cls_name}.create() first.")

        return cls.instance

    @classmethod
    def create(cls, mode: str, mem_size: int) -> MemCacheHandlerBase:
        """Create a new MemCacheHandlerBase instance.

        Args:
            mode (str): There are two options: null, multiprocessing or singleprocessing.
            mem_size (int): The size of memory pool (bytes).
        """
        logger.info(f"Try to create a {mem_size} size memory pool.")

        _, world_size = get_dist_info()
        if world_size > 1:
            mem_size = mem_size // world_size
            logger.info(f"Since world_size={world_size} > 1, each worker a {mem_size} size memory pool.")

        if mode == "null" or mem_size == 0:
            cls.instance = MemCacheHandlerBase(mem_size=0)
            cls.instance.freeze()
        elif mode == "multiprocessing":
            cls.instance = MemCacheHandlerForMP(mem_size)
        elif mode == "singleprocessing":
            cls.instance = MemCacheHandlerForSP(mem_size)
        else:
            raise MemCacheHandlerError(f"{mode} is unknown mode.")

        return cls.instance

    @classmethod
    def delete(cls) -> None:
        """Delete the existing MemCacheHandlerBase instance."""
        if hasattr(cls, "instance"):
            del cls.instance



"""Memory cache hook for logging and freezing MemCacheHandler."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmcv.runner.hooks import HOOKS, Hook

from .mem_cache_handler import MemCacheHandlerSingleton


@HOOKS.register_module()
class MemCacheHook(Hook):
    """Memory cache hook for logging and freezing MemCacheHandler."""

    def __init__(self) -> None:
        self.handler = MemCacheHandlerSingleton.get()
        # It is because the first evaluation comes at the very beginning of the training.
        # We don't want to cache validation samples first.
        self.handler.freeze()

    def before_epoch(self, runner):
        """Before training, unfreeze the handler."""
        # We want to cache training samples first.
        self.handler.unfreeze()

    def after_epoch(self, runner):
        """After epoch. Log the handler statistics.

        To prevent it from skipping the validation samples,
        this hook should have lower priority than CustomEvalHook.
        """
        self.handler.freeze()
        runner.logger.info(f"{self.handler}")



"""Module defining Mix-in class of SAMClassifier."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import datumaro as dm
import numpy as np
import pandas as pd

from otx.algorithms.common.utils.logger import get_logger
from otx.api.entities.dataset_item import DatasetItemEntityWithID
from otx.api.entities.datasets import DatasetEntity
from otx.core.data.noisy_label_detection import LossDynamicsTracker, LossDynamicsTrackingMixin

logger = get_logger()


class SAMClassifierMixin:
    """SAM-enabled BaseClassifier mix-in."""

    def train_step(self, data, optimizer=None, **kwargs):
        """Saving current batch data to compute SAM gradient."""
        self.current_batch = data
        return super().train_step(data, optimizer, **kwargs)


class MultiClassClsLossDynamicsTracker(LossDynamicsTracker):
    """Loss dynamics tracker for multi-class classification task."""

    def __init__(self) -> None:
        super().__init__()

    def init_with_otx_dataset(self, otx_dataset: DatasetEntity[DatasetItemEntityWithID]) -> None:
        """DatasetEntity should be injected to the tracker for the initialization."""
        otx_labels = otx_dataset.get_labels()
        label_categories = dm.LabelCategories.from_iterable([label_entity.name for label_entity in otx_labels])
        self.otx_label_map = {label_entity.id_: idx for idx, label_entity in enumerate(otx_labels)}

        def _convert_anns(item: DatasetItemEntityWithID):
            labels = [
                dm.Label(label=self.otx_label_map[label.id_])
                for ann in item.get_annotations()
                for label in ann.get_labels()
            ]
            return labels

        self._export_dataset = dm.Dataset.from_iterable(
            [
                dm.DatasetItem(
                    id=item.id_,
                    subset="train",
                    media=dm.Image.from_file(path=item.media.path, size=(item.media.height, item.media.width))
                    if item.media.path
                    else dm.Image.from_numpy(
                        data=getattr(item.media, "_Image__data"), size=(item.media.height, item.media.width)
                    ),
                    annotations=_convert_anns(item),
                )
                for item in otx_dataset
            ],
            infos={"purpose": "noisy_label_detection", "task": "OTX-MultiClassCls"},
            categories={dm.AnnotationType.label: label_categories},
        )

        super().init_with_otx_dataset(otx_dataset)

    def accumulate(self, outputs, iter) -> None:
        """Accumulate training loss dynamics for each training step."""
        entity_ids = outputs["entity_ids"]
        label_ids = np.squeeze(outputs["label_ids"])
        loss_dyns = outputs["loss_dyns"]

        for entity_id, label_id, loss_dyn in zip(entity_ids, label_ids, loss_dyns):
            self._loss_dynamics[(entity_id, label_id)].append((iter, loss_dyn))

    def export(self, output_path: str) -> None:
        """Export loss dynamics statistics to Datumaro format."""
        df = pd.DataFrame.from_dict(
            {
                k: (np.array([iter for iter, _ in arr]), np.array([value for _, value in arr]))
                for k, arr in self._loss_dynamics.items()
            },
            orient="index",
            columns=["iters", "loss_dynamics"],
        )

        for (entity_id, label_id), row in df.iterrows():
            item = self._export_dataset.get(entity_id, "train")
            for ann in item.annotations:
                if isinstance(ann, dm.Label) and ann.label == self.otx_label_map[label_id]:
                    ann.attributes = row.to_dict()

        self._export_dataset.export(output_path, format="datumaro")


class ClsLossDynamicsTrackingMixin(LossDynamicsTrackingMixin):
    """Mix-in to track loss dynamics during training for classification tasks."""

    def __init__(self, track_loss_dynamics: bool = False, **kwargs):
        if track_loss_dynamics:
            if getattr(self, "multilabel", False) or getattr(self, "hierarchical", False):
                raise RuntimeError("multilabel or hierarchical tasks are not supported now.")

            head_cfg = kwargs.get("head", None)
            loss_cfg = head_cfg.get("loss", None)
            loss_cfg["reduction"] = "none"

        # This should be called after modifying "reduction" config.
        super().__init__(**kwargs)

        # This should be called after super().__init__(),
        # since LossDynamicsTrackingMixin.__init__() creates self._loss_dyns_tracker
        self._loss_dyns_tracker = MultiClassClsLossDynamicsTracker()

    def train_step(self, data, optimizer=None, **kwargs):
        """The iteration step for training.

        If self._track_loss_dynamics = False, just follow BaseClassifier.train_step().
        Otherwise, it steps with tracking loss dynamics.
        """
        if self.loss_dyns_tracker.initialized:
            return self._train_step_with_tracking(data, optimizer, **kwargs)
        return super().train_step(data, optimizer, **kwargs)

    def _train_step_with_tracking(self, data, optimizer=None, **kwargs):
        losses = self(**data)

        loss_dyns = losses["loss"].detach().cpu().numpy()
        assert not np.isscalar(loss_dyns)

        entity_ids = [img_meta["entity_id"] for img_meta in data["img_metas"]]
        label_ids = [img_meta["label_id"] for img_meta in data["img_metas"]]
        loss, log_vars = self._parse_losses(losses)

        outputs = dict(
            loss=loss,
            log_vars=log_vars,
            loss_dyns=loss_dyns,
            entity_ids=entity_ids,
            label_ids=label_ids,
            num_samples=len(data["img"].data),
        )

        return outputs



"""Hook module to track loss dynamics during training and export these statistics to Datumaro format."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import os.path as osp

from mmcv.parallel import MMDataParallel
from mmcv.runner import BaseRunner
from mmcv.runner.hooks import HOOKS, Hook
from mmcv.utils import Config, ConfigDict

from otx.algorithms.common.adapters.mmcv.utils.config_utils import (
    update_or_add_custom_hook,
)
from otx.algorithms.common.utils.logger import get_logger
from otx.api.entities.datasets import DatasetEntity

from .base import LossDynamicsTracker, LossDynamicsTrackingMixin

logger = get_logger()

__all__ = ["LossDynamicsTrackingHook"]


@HOOKS.register_module()
class LossDynamicsTrackingHook(Hook):
    """Tracking loss dynamics during training and export it to Datumaro dataset format."""

    def __init__(self, output_path: str, alpha: float = 0.001) -> None:
        self._output_fpath = osp.join(output_path, "noisy_label_detection")

    def before_run(self, runner):
        """Before run, check the type of model for safe running."""
        if not isinstance(runner.model, MMDataParallel):
            raise NotImplementedError(f"Except MMDataParallel, runner.model={type(runner.model)} is not supported now.")

    def before_train_epoch(self, runner: BaseRunner):
        """Initialize the tracker for training loss dynamics tracking.

        Tracker needs the training dataset for initialization.
        However, there is no way to access to dataloader until the beginning of training epoch.
        """
        self._init_tracker(runner, runner.data_loader.dataset.otx_dataset)

    def _get_tracker(self, runner: BaseRunner) -> LossDynamicsTracker:
        model = runner.model.module

        if not isinstance(model, LossDynamicsTrackingMixin):
            raise RuntimeError(
                f"The model should be an instance of LossDynamicsTrackingMixin, but type(model)={type(model)}."
            )
        return model.loss_dyns_tracker

    def _init_tracker(self, runner: BaseRunner, otx_dataset: DatasetEntity) -> None:
        tracker = self._get_tracker(runner)
        if tracker.initialized:
            return

        logger.info("Initialize training loss dynamics tracker.")
        tracker.init_with_otx_dataset(otx_dataset)

    def after_train_iter(self, runner):
        """Accumulate training loss dynamics.

        It should be here because it needs to access the training iteration.
        """
        tracker = self._get_tracker(runner)
        tracker.accumulate(runner.outputs, runner.iter)

    def after_run(self, runner: BaseRunner) -> None:
        """Export loss dynamics statistics to Datumaro format."""

        tracker = self._get_tracker(runner)

        if tracker.initialized:
            logger.info(f"Export training loss dynamics to {self._output_fpath}")
            tracker.export(self._output_fpath)

    @classmethod
    def configure_recipe(cls, recipe_cfg: Config, output_path: str) -> None:
        """Configure recipe to enable loss dynamics tracking."""
        recipe_cfg.model["track_loss_dynamics"] = True

        update_or_add_custom_hook(
            recipe_cfg,
            ConfigDict(
                type="LossDynamicsTrackingHook",
                priority="LOWEST",
                output_path=output_path,
            ),
        )



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from unittest.mock import patch

import numpy as np
import pytest
from openvino.model_zoo.model_api.models.utils import Detection

from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.id import ID
from otx.api.entities.label import Color, Domain, LabelEntity
from otx.api.entities.label_schema import LabelGroup, LabelSchemaEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.usecases.exportable_code.prediction_to_annotation_converter import (
    AnomalyClassificationToAnnotationConverter,
    AnomalyDetectionToAnnotationConverter,
    AnomalySegmentationToAnnotationConverter,
    ClassificationToAnnotationConverter,
    DetectionBoxToAnnotationConverter,
    DetectionToAnnotationConverter,
    IPredictionToAnnotationConverter,
    SegmentationToAnnotationConverter,
    create_converter,
)
from otx.api.utils.time_utils import now
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestDetectionToAnnotationConverter:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_detection_to_annotation_convert(self):
        """
        <b>Description:</b>
        Check that DetectionToAnnotationConverter correctly converts Network output to list of Annotation

        <b>Input data:</b>
        Array of network output with shape [4,6]

        <b>Expected results:</b>
        Test passes if each Converted annotation has the same  values as the network output

        <b>Steps</b>
        1. Create mock network output
        2. Convert network output to Annotation
        3. Check Annotations
        """
        test_boxes = np.array(
            (
                (0, 0.6, 0.1, 0.1, 0.2, 0.3),
                (1, 0.2, 0.2, 0.1, 0.3, 0.4),
                (1, 0.7, 0.3, 0.2, 0.5, 0.6),
                (0, 0.1, 0.1, 0.1, 0.2, 0.3),
            )
        )

        labels = [
            LabelEntity("Zero", domain=Domain.DETECTION),
            LabelEntity("One", domain=Domain.DETECTION),
        ]

        converter = DetectionToAnnotationConverter(labels)

        annotation_scene = converter.convert_to_annotation(test_boxes)

        for i, annotation in enumerate(annotation_scene.annotations):
            label: ScoredLabel = next(iter(annotation.get_labels()))
            test_label = labels[int(test_boxes[i][0])]
            assert test_label.name == label.name

            assert test_boxes[i][1], label.probability

            assert test_boxes[i][2] == annotation.shape.x1
            assert test_boxes[i][3] == annotation.shape.y1
            assert test_boxes[i][4] == annotation.shape.x2
            assert test_boxes[i][5] == annotation.shape.y2

        annotation_scene = converter.convert_to_annotation(np.ndarray((0, 6)))
        assert 0 == len(annotation_scene.shapes)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_detection_to_annotation_convert_openvino_shape(self):
        """
        <b>Description:</b>
        Check that DetectionToAnnotationConverter correctly converts OpenVINO Network output to annotations

        <b>Input data:</b>
        Array of network output with shape [4,7]

        <b>Expected results:</b>
        Test passes if each Converted annotation has the same values as the network output

        <b>Steps</b>
        1. Create mock network output
        2. Convert network output to Annotation
        3. Check Annotations
        """
        test_boxes = np.array(
            (
                (-12, 0, 0.6, 0.1, 0.1, 0.2, 0.3),
                (12, 1, 0.2, 0.0, 0.1, 0.1, 0.2),
                (1234, 1, 0.7, 0.2, 0.4, 0.7, 0.5),
                (1251, 0, 0.1, 0.1, 0.1, 0.2, 0.3),
            )
        )

        labels = [
            LabelEntity("Zero", domain=Domain.DETECTION),
            LabelEntity("One", domain=Domain.DETECTION),
        ]

        converter = DetectionToAnnotationConverter(labels)

        annotation_scene = converter.convert_to_annotation(test_boxes)

        for i, annotation in enumerate(annotation_scene.annotations):
            label: ScoredLabel = next(iter(annotation.get_labels()))
            test_label = labels[int(test_boxes[i][1])]
            assert test_label.name == label.name

            assert test_boxes[i][2] == label.probability

            assert test_boxes[i][3] == annotation.shape.x1
            assert test_boxes[i][4] == annotation.shape.y1
            assert test_boxes[i][5] == annotation.shape.x2
            assert test_boxes[i][6] == annotation.shape.y2

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_detection_to_annotation_convert_invalid_input(self):
        """
        <b>Description:</b>
        Check that DetectionToAnnotationConverter raises an error if invalid inputs are provided

        <b>Input data:</b>
        Array of size [1203, 5]
        Array of size [3, 8]

        <b>Expected results:</b>
        Test passes a ValueError is raised for both inputs

        <b>Steps</b>
        1. Create DetectionToAnnotationConverter
        2. Attempt to convert array of [1203,5] to annotations
        3. Attempt to convert array of [3, 8] to annotations
        """
        labels = [
            LabelEntity("Zero", domain=Domain.DETECTION),
            LabelEntity("One", domain=Domain.DETECTION),
        ]
        converter = DetectionToAnnotationConverter(labels)

        with pytest.raises(ValueError):
            converter.convert_to_annotation(np.ndarray((1203, 5)))

        with pytest.raises(ValueError):
            converter.convert_to_annotation(np.ndarray((3, 8)))


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestIPredictionToAnnotation:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    @patch(
        (
            "otx.api.usecases.exportable_code.prediction_to_annotation_converter."
            "IPredictionToAnnotationConverter.__abstractmethods__"
        ),
        set(),
    )
    def test_i_prediction_to_annotation(self):
        """
        <b>Description:</b>
        Check "IPredictionToAnnotationConverter" class "convert_to_annotation" method

        <b>Input data:</b>
        "IPredictionToAnnotationConverter" class object, "predictions" array, "metadata" dictionary

        <b>Expected results:</b>
        Test passes if "NotImplementedError" exception is raised by "convert_to_annotation" method
        """
        i_prediction_to_annotation_converter = IPredictionToAnnotationConverter()
        with pytest.raises(NotImplementedError):
            i_prediction_to_annotation_converter.convert_to_annotation(
                predictions=np.random.randint(low=0, high=255, size=(5, 5, 3)),
                metadata={},
            )


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestCreateConverter:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_create_converter(self):
        """
        <b>Description:</b>
        Check "create_converter" function

        <b>Input data:</b>
        "converter_type" Domain-class object, "labels" LabelSchemaEntity-class object

        <b>Expected results:</b>
        Test passes if "IPredictionToAnnotationConverter" object returned by "create_converter" function is equal
        to expected

        <b>Steps</b>
        1. Check "DetectionBoxToAnnotationConverter" object returned by "create_converter" function when
        "DETECTION" domain is specified as "converter_type" parameter
        2. Check "SegmentationToAnnotationConverter" object returned by "create_converter" function when
        "SEGMENTATION" domain is specified as "converter_type" parameter
        3. Check "ClassificationToAnnotationConverter" object returned by "create_converter" function when
        "CLASSIFICATION" domain is specified as "converter_type" parameter
        4. Check "AnomalyClassificationToAnnotationConverter" object returned by "create_converter" function when
        "ANOMALY_CLASSIFICATION" domain is specified as "converter_type" parameter
        5. Check that "ValueError" exception is raised when "ANOMALY_DETECTION" or "ANOMALY_SEGMENTATION" domain is
        specified as "converter_type" parameter
        """
        # Checking "DetectionBoxToAnnotationConverter" returned by "create_converter" function when "DETECTION" is
        # specified as "converter_type"
        labels = [
            LabelEntity(name="Detection label", domain=Domain.DETECTION, id=ID("1")),
            LabelEntity(name="Other Detection label", domain=Domain.DETECTION, id=ID("2")),
        ]
        label_group = LabelGroup(name="Detection labels group", labels=labels)
        label_schema = LabelSchemaEntity(label_groups=[label_group])
        converter = create_converter(converter_type=Domain.DETECTION, labels=label_schema)
        assert isinstance(converter, DetectionToAnnotationConverter)
        assert converter.labels == labels
        # Checking "SegmentationToAnnotationConverter" returned by "create_converter" function when "SEGMENTATION"is
        # specified as "converter_type"
        labels = [
            LabelEntity(name="Segmentation label", domain=Domain.SEGMENTATION, id=ID("1")),
            LabelEntity(
                name="Other Segmentation label",
                domain=Domain.SEGMENTATION,
                id=ID("2"),
            ),
        ]
        label_group = LabelGroup(name="Segmentation labels group", labels=labels)
        label_schema = LabelSchemaEntity(label_groups=[label_group])
        converter = create_converter(converter_type=Domain.SEGMENTATION, labels=label_schema)
        assert isinstance(converter, SegmentationToAnnotationConverter)
        assert converter.label_map == {1: labels[0], 2: labels[1]}
        # Checking "ClassificationToAnnotationConverter" returned by "create_converter" function when
        # "CLASSIFICATION" is specified as "converter_type"
        labels = [
            LabelEntity(
                name="Classification label",
                domain=Domain.CLASSIFICATION,
                id=ID("1"),
            ),
            LabelEntity(
                name="Other Classification label",
                domain=Domain.CLASSIFICATION,
                id=ID("2"),
            ),
        ]
        label_group = LabelGroup(name="Classification labels group", labels=labels)
        label_schema = LabelSchemaEntity(label_groups=[label_group])
        converter = create_converter(converter_type=Domain.CLASSIFICATION, labels=label_schema)
        assert isinstance(converter, ClassificationToAnnotationConverter)
        assert converter.labels == labels
        # Checking that "AnomalyClassificationToAnnotationConverter" returned by "create_converter" function when
        # "ANOMALY_CLASSIFICATION" is specified as "converter_type"
        labels = [
            LabelEntity(name="Normal", domain=Domain.ANOMALY_CLASSIFICATION, id=ID("1")),
            LabelEntity(
                name="Anomalous",
                domain=Domain.ANOMALY_CLASSIFICATION,
                id=ID("2"),
                is_anomalous=True,
            ),
        ]
        label_group = LabelGroup(name="Anomaly classification labels group", labels=labels)
        label_schema = LabelSchemaEntity(label_groups=[label_group])
        converter = create_converter(converter_type=Domain.ANOMALY_CLASSIFICATION, labels=label_schema)
        assert isinstance(converter, AnomalyClassificationToAnnotationConverter)
        assert converter.normal_label == labels[0]
        assert converter.anomalous_label == labels[1]
        # Checking that "AnomalyDetectionToAnnotationConverter" returned by "create_converter" function when
        # "ANOMALY_DETECTION" is specified as "converter_type"
        labels = [
            LabelEntity(name="Normal", domain=Domain.ANOMALY_DETECTION, id=ID("1")),
            LabelEntity(
                name="Anomalous",
                domain=Domain.ANOMALY_DETECTION,
                id=ID("2"),
                is_anomalous=True,
            ),
        ]
        label_group = LabelGroup(name="Anomaly detection labels group", labels=labels)
        label_schema = LabelSchemaEntity(label_groups=[label_group])
        converter = create_converter(converter_type=Domain.ANOMALY_DETECTION, labels=label_schema)
        assert isinstance(converter, AnomalyDetectionToAnnotationConverter)
        assert converter.normal_label == labels[0]
        assert converter.anomalous_label == labels[1]
        # Checking that "AnomalySegmentationToAnnotationConverter" returned by "create_converter" function when
        # "ANOMALY_SEGMENTATION" is specified as "converter_type"
        labels = [
            LabelEntity(name="Normal", domain=Domain.ANOMALY_SEGMENTATION, id=ID("1")),
            LabelEntity(
                name="Anomalous",
                domain=Domain.ANOMALY_SEGMENTATION,
                id=ID("2"),
                is_anomalous=True,
            ),
        ]
        label_group = LabelGroup(name="Anomaly detection labels group", labels=labels)
        label_schema = LabelSchemaEntity(label_groups=[label_group])
        converter = create_converter(converter_type=Domain.ANOMALY_SEGMENTATION, labels=label_schema)
        assert isinstance(converter, AnomalySegmentationToAnnotationConverter)
        assert converter.normal_label == labels[0]
        assert converter.anomalous_label == labels[1]


def check_annotation_scene(annotation_scene: AnnotationSceneEntity, expected_length: int):
    assert isinstance(annotation_scene, AnnotationSceneEntity)
    assert annotation_scene.kind == AnnotationSceneKind.PREDICTION
    assert len(annotation_scene.annotations) == expected_length


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestDetectionBoxToAnnotation:
    color = Color(red=180, green=230, blue=30)
    creation_date = now()
    labels = [
        LabelEntity(
            name="Detection label",
            domain=Domain.DETECTION,
            color=color,
            creation_date=creation_date,
            id=ID("1"),
        ),
        LabelEntity(
            name="Other Detection label",
            domain=Domain.DETECTION,
            color=color,
            creation_date=creation_date,
            id=ID("2"),
        ),
    ]

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_detection_box_to_annotation_init(self):
        """
        <b>Description:</b>
        Check "DetectionBoxToAnnotationConverter" class object initialization

        <b>Input data:</b>
        "DetectionBoxToAnnotationConverter" class object with specified "labels" parameter

        <b>Expected results:</b>
        Test passes if attributes of initialized "DetectionBoxToAnnotationConverter" object are equal to expected

        <b>Steps</b>
        1. Check "labels" attribute of "DetectionBoxToAnnotationConverter" object initialized with non-empty labels
        list
        2. Check "labels" attribute of "DetectionBoxToAnnotationConverter" object initialized with empty labels list
        3. Check "labels" attributes of "DetectionBoxToAnnotationConverter" object initialized with non-empty and
        empty labels list
        """
        # Checking "labels" of "DetectionBoxToAnnotationConverter" initialized with non-empty labels list
        non_empty_labels = self.labels
        label_group = LabelGroup(name="Detection labels group", labels=non_empty_labels)
        label_schema = LabelSchemaEntity(label_groups=[label_group])
        converter = DetectionBoxToAnnotationConverter(labels=label_schema)
        assert converter.labels == non_empty_labels
        # Checking "labels" of "DetectionBoxToAnnotationConverter" initialized with empty labels list
        empty_labels = [
            LabelEntity(
                name="empty label",
                domain=Domain.DETECTION,
                is_empty=True,
                id=ID("3"),
            ),
            LabelEntity(
                name="other empty label",
                domain=Domain.DETECTION,
                is_empty=True,
                id=ID("4"),
            ),
        ]
        label_group = LabelGroup(name="Detection labels group", labels=empty_labels)
        label_schema = LabelSchemaEntity(label_groups=[label_group])
        converter = DetectionBoxToAnnotationConverter(labels=label_schema)
        assert converter.labels == []
        # Checking "labels" of "DetectionBoxToAnnotationConverter" initialized with non-empty and empty labels list
        label_group = LabelGroup(name="Detection labels group", labels=non_empty_labels + empty_labels)
        label_schema = LabelSchemaEntity(label_groups=[label_group])
        converter = DetectionBoxToAnnotationConverter(labels=label_schema)
        assert converter.labels == non_empty_labels

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_detection_box_to_annotation_convert(self):
        """
        <b>Description:</b>
        Check "DetectionBoxToAnnotationConverter" class "convert_to_annotation" method

        <b>Input data:</b>
        "DetectionBoxToAnnotationConverter" class object, "predictions" list with Detection-class objects,
        "metadata" dictionary

        <b>Expected results:</b>
        Test passes if "AnnotationSceneEntity" object returned by "convert_to_annotation" method is equal to
        expected
        """

        def check_annotation(
            actual_annotation: Annotation,
            expected_label: LabelEntity,
            expected_probability: float,
            expected_x1: float,
            expected_y1: float,
            expected_x2: float,
            expected_y2: float,
        ):
            assert isinstance(actual_annotation, Annotation)
            assert actual_annotation.get_labels() == [
                ScoredLabel(label=expected_label, probability=expected_probability)
            ]
            assert isinstance(actual_annotation.shape, Rectangle)
            assert actual_annotation.shape.x1 == pytest.approx(expected_x1)
            assert actual_annotation.shape.y1 == pytest.approx(expected_y1)
            assert actual_annotation.shape.x2 == pytest.approx(expected_x2)
            assert actual_annotation.shape.y2 == pytest.approx(expected_y2)

        labels = self.labels
        label_group = LabelGroup(name="Detection labels group", labels=labels)
        label_schema = LabelSchemaEntity(label_groups=[label_group])
        converter = DetectionBoxToAnnotationConverter(labels=label_schema)
        metadata = {
            "non-required key": 1,
            "other non-required key": 2,
            "original_shape": [10, 20],
        }
        box_1 = Detection(xmin=2, ymin=2, xmax=4, ymax=6, score=0.8, id=0)
        box_2 = Detection(xmin=6, ymin=4, xmax=10, ymax=9, score=0.9, id=1)
        predictions = [box_1, box_2]
        predictions_to_annotations = converter.convert_to_annotation(predictions=predictions, metadata=metadata)
        check_annotation_scene(annotation_scene=predictions_to_annotations, expected_length=2)
        check_annotation(
            actual_annotation=predictions_to_annotations.annotations[0],
            expected_label=labels[0],
            expected_probability=0.8,
            expected_x1=0.1,
            expected_y1=0.2,
            expected_x2=0.2,
            expected_y2=0.6,
        )
        check_annotation(
            actual_annotation=predictions_to_annotations.annotations[1],
            expected_label=labels[1],
            expected_probability=0.9,
            expected_x1=0.3,
            expected_y1=0.4,
            expected_x2=0.5,
            expected_y2=0.9,
        )


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestSegmentationToAnnotation:
    color = Color(red=180, green=230, blue=30)
    creation_date = now()
    labels = [
        LabelEntity(
            name="Segmentation label",
            domain=Domain.SEGMENTATION,
            color=color,
            creation_date=creation_date,
            id=ID("0"),
        ),
        LabelEntity(
            name="Other Segmentation label",
            domain=Domain.SEGMENTATION,
            color=color,
            creation_date=creation_date,
            id=ID("1"),
        ),
    ]

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_segmentation_to_annotation_init(self):
        """
        <b>Description:</b>
        Check "SegmentationToAnnotationConverter" class object initialization

        <b>Input data:</b>
        "SegmentationToAnnotationConverter" class object with specified "label_schema" parameter

        <b>Expected results:</b>
        Test passes if attributes of initialized "SegmentationToAnnotationConverter" object are equal to expected

        <b>Steps</b>
        1. Check "label_map" attribute of "SegmentationToAnnotationConverter" object initialized with non-empty
        labels list
        2. Check "label_map" attribute of "SegmentationToAnnotationConverter" object initialized with empty labels
        list
        3. Check "label_map" attributes of "SegmentationToAnnotationConverter" object initialized with non-empty and
        empty labels list
        """
        # Checking "label_map" of "SegmentationToAnnotationConverter" initialized with non-empty labels list
        non_empty_labels = self.labels
        label_group = LabelGroup(name="Segmentation labels group", labels=non_empty_labels)
        label_schema = LabelSchemaEntity(label_groups=[label_group])
        converter = SegmentationToAnnotationConverter(label_schema=label_schema)
        expected_non_empty_labels_map = {
            1: non_empty_labels[0],
            2: non_empty_labels[1],
        }
        assert converter.label_map == expected_non_empty_labels_map
        # Checking "label_map" of "SegmentationToAnnotationConverter" initialized with empty labels list
        empty_labels = [
            LabelEntity(
                name="empty label",
                domain=Domain.SEGMENTATION,
                is_empty=True,
                id=ID("3"),
            ),
            LabelEntity(
                name="other empty label",
                domain=Domain.SEGMENTATION,
                is_empty=True,
                id=ID("4"),
            ),
        ]
        label_group = LabelGroup(name="Segmentation labels group", labels=empty_labels)
        label_schema = LabelSchemaEntity(label_groups=[label_group])
        converter = SegmentationToAnnotationConverter(label_schema=label_schema)
        assert converter.label_map == {}
        # Checking "label_map" of "SegmentationToAnnotationConverter" initialized with non-empty and empty labels list
        label_group = LabelGroup(name="Segmentation labels group", labels=non_empty_labels + empty_labels)
        label_schema = LabelSchemaEntity(label_groups=[label_group])
        converter = SegmentationToAnnotationConverter(label_schema=label_schema)
        assert converter.label_map == expected_non_empty_labels_map

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_segmentation_to_annotation_convert(self):
        """
        <b>Description:</b>
        Check "SegmentationToAnnotationConverter" class "convert_to_annotation" method

        <b>Input data:</b>
        "SegmentationToAnnotationConverter" class object, "predictions" array with hard predictions,
        "metadata" dictionary

        <b>Expected results:</b>
        Test passes if "AnnotationSceneEntity" object returned by "convert_to_annotation" method is equal to
        expected
        """

        def check_annotation(
            actual_annotation: Annotation,
            expected_label: LabelEntity,
            expected_probability: float,
            expected_points: list,
        ):
            assert isinstance(actual_annotation, Annotation)
            annotation_labels = actual_annotation.get_labels()
            # Checking Annotation ScoredLabel
            assert len(annotation_labels) == 1
            assert isinstance(annotation_labels[0], ScoredLabel)
            assert annotation_labels[0].label == expected_label
            assert annotation_labels[0].probability == pytest.approx(expected_probability)
            # Checking Annotation Shape
            assert isinstance(actual_annotation.shape, Polygon)
            assert actual_annotation.shape.points == expected_points

        labels = self.labels
        label_group = LabelGroup(name="Segmentation labels group", labels=labels)
        label_schema = LabelSchemaEntity(label_groups=[label_group])
        converter = SegmentationToAnnotationConverter(label_schema=label_schema)
        soft_prediction = np.array(
            [
                (
                    [0.8, 0.1, 0.2],
                    [0.9, 0.1, 0.2],
                    [0.3, 0.2, 0.8],
                    [0.1, 0.2, 0.8],
                ),
                (
                    [0.1, 0.8, 0.3],
                    [0.0, 0.8, 0.2],
                    [0.2, 0.1, 0.9],
                    [0.0, 0.2, 0.8],
                ),
                (
                    [0.1, 0.7, 0.3],
                    [0.3, 0.8, 0.2],
                    [0.1, 0.2, 0.8],
                    [0.4, 0.3, 0.7],
                ),
                (
                    [0.0, 1.0, 0.0],
                    [0.1, 0.9, 0.1],
                    [0.1, 0.1, 0.9],
                    [0.2, 0.2, 0.8],
                ),
            ]
        )
        hard_predictions = np.array([(0, 0, 2, 2), (1, 1, 2, 2), (1, 1, 2, 2), (1, 1, 2, 2)])

        metadata = {
            "non-required key": 1,
            "other non-required key": 2,
            "soft_prediction": soft_prediction,
        }

        predictions_to_annotations = converter.convert_to_annotation(predictions=hard_predictions, metadata=metadata)
        check_annotation_scene(annotation_scene=predictions_to_annotations, expected_length=2)
        check_annotation(
            actual_annotation=predictions_to_annotations.annotations[0],
            expected_label=labels[0],
            expected_probability=0.8333333333333333,
            expected_points=[
                Point(0.0, 0.25),
                Point(0.0, 0.5),
                Point(0.0, 0.75),
                Point(0.25, 0.75),
                Point(0.25, 0.5),
                Point(0.25, 0.25),
            ],
        )
        check_annotation(
            actual_annotation=predictions_to_annotations.annotations[1],
            expected_label=labels[1],
            expected_probability=0.8125,
            expected_points=[
                Point(0.5, 0.0),
                Point(0.5, 0.25),
                Point(0.5, 0.5),
                Point(0.5, 0.75),
                Point(0.75, 0.75),
                Point(0.75, 0.5),
                Point(0.75, 0.25),
                Point(0.75, 0.0),
            ],
        )

    @pytest.mark.components(OtxSdkComponent.OTX_API)
    class TestClassificationToAnnotation:
        @pytest.mark.priority_medium
        @pytest.mark.unit
        @pytest.mark.reqids(Requirements.REQ_1)
        def test_classification_to_annotation_init(self):
            """
            <b>Description:</b>
            Check "ClassificationToAnnotationConverter" class object initialization

            <b>Input data:</b>
            "ClassificationToAnnotationConverter" class object with specified "label_schema" parameter

            <b>Expected results:</b>
            Test passes if attributes of initialized "ClassificationToAnnotationConverter" object are equal to
            expected

            <b>Steps</b>
            1. Check attributes of "ClassificationToAnnotationConverter" object initialized with one label group
            with non-empty labels list length more than 1
            2. Check attributes of "ClassificationToAnnotationConverter" object initialized with one label group
            with non-empty labels list length equal to 1
            3. Check attributes of "ClassificationToAnnotationConverter" object initialized with two label groups
            with one label in each
            4. Check attributes of "ClassificationToAnnotationConverter" object initialized with two label groups
            with several labels in each
            """
            label_0 = LabelEntity(name="label_0", domain=Domain.CLASSIFICATION, id=ID("0"))
            label_0_1 = LabelEntity(name="label_0_1", domain=Domain.CLASSIFICATION, id=ID("0_1"))
            label_0_2 = LabelEntity(name="label_0_2", domain=Domain.CLASSIFICATION, id=ID("0_2"))
            label_0_1_1 = LabelEntity(name="label_0_1_1", domain=Domain.CLASSIFICATION, id=ID("0_1_1"))

            non_empty_labels = [label_0, label_0_1, label_0_1_1, label_0_2]
            empty_labels = [
                LabelEntity(
                    name="empty label",
                    domain=Domain.CLASSIFICATION,
                    is_empty=True,
                    id=ID("3"),
                )
            ]
            # Checking attributes of "ClassificationToAnnotationConverter" initialized with one label group with
            # non-empty labels list length more than 1
            label_group = LabelGroup(
                name="Classification labels group",
                labels=non_empty_labels + empty_labels,
            )
            label_schema = LabelSchemaEntity(label_groups=[label_group])
            converter = ClassificationToAnnotationConverter(label_schema=label_schema)
            assert converter.labels == non_empty_labels
            assert converter.empty_label == empty_labels[0]
            assert converter.label_schema == label_schema
            assert not converter.hierarchical
            # Checking attributes of "ClassificationToAnnotationConverter" initialized with one label group with
            # non-empty labels list length equal to 1
            label_group = LabelGroup(name="Classification labels group", labels=[label_0] + empty_labels)
            label_schema = LabelSchemaEntity(label_groups=[label_group])
            converter = ClassificationToAnnotationConverter(label_schema=label_schema)
            assert converter.labels == [label_0] + empty_labels
            assert converter.empty_label == empty_labels[0]
            assert converter.label_schema == label_schema
            assert not converter.hierarchical
            # Checking attributes of "ClassificationToAnnotationConverter" initialized with two label groups with
            # one label in each
            label_group = LabelGroup(name="Classification labels group", labels=[label_0_1])
            other_label_group = LabelGroup(name="Other Classification labels group", labels=[label_0_2])
            label_schema = LabelSchemaEntity(label_groups=[label_group, other_label_group])
            converter = ClassificationToAnnotationConverter(label_schema=label_schema)
            assert converter.labels == [label_0_1, label_0_2]
            assert not converter.empty_label
            assert converter.label_schema == label_schema
            assert not converter.hierarchical
            # Checking attributes of "ClassificationToAnnotationConverter" initialized with two label groups with
            # several labels in each
            other_non_empty_labels = [
                LabelEntity(name="label", domain=Domain.CLASSIFICATION, id=ID("3")),
                LabelEntity(name="other label", domain=Domain.CLASSIFICATION, id=ID("4")),
            ]
            label_group = LabelGroup(name="Classification labels group", labels=non_empty_labels)
            other_label_group = LabelGroup(
                name="Other Classification labels group",
                labels=other_non_empty_labels,
            )
            label_schema = LabelSchemaEntity(label_groups=[label_group, other_label_group])
            converter = ClassificationToAnnotationConverter(label_schema=label_schema)
            assert converter.labels == non_empty_labels + other_non_empty_labels
            assert not converter.empty_label
            assert converter.label_schema == label_schema
            assert converter.hierarchical

        @pytest.mark.priority_medium
        @pytest.mark.unit
        @pytest.mark.reqids(Requirements.REQ_1)
        def test_classification_to_annotation_convert(self):
            """
            <b>Description:</b>
            Check "ClassificationToAnnotationConverter" class "convert_to_annotation" method

            <b>Input data:</b>
            "ClassificationToAnnotationConverter" class object, "predictions" list, "metadata" dictionary

            <b>Expected results:</b>
            Test passes if "AnnotationSceneEntity" object returned by "convert_to_annotation" method is equal to
            expected

            <b>Steps</b>
            1. Check attributes of "AnnotationSceneEntity" object returned by "convert_to_annotation" method for
            "ClassificationToAnnotationConverter" object initialized with label group with several non-empty labels
            2. Check attributes of "AnnotationSceneEntity" object returned by "convert_to_annotation" method with
            "predictions" parameter equal to empty list
            3. Check attributes of "AnnotationSceneEntity" object returned by "convert_to_annotation" method for
            "ClassificationToAnnotationConverter" object initialized with several LabelGroups
            """

            def check_annotation(actual_annotation: Annotation, expected_labels: list):
                assert isinstance(actual_annotation, Annotation)
                assert actual_annotation.get_labels(include_empty=True) == expected_labels
                assert isinstance(actual_annotation.shape, Rectangle)
                assert Rectangle.is_full_box(rectangle=actual_annotation.shape)

            label_0 = LabelEntity(name="label_0", domain=Domain.CLASSIFICATION, id=ID("0"))
            label_0_1 = LabelEntity(name="label_0_1", domain=Domain.CLASSIFICATION, id=ID("0_1"))
            label_0_2 = LabelEntity(name="label_0_2", domain=Domain.CLASSIFICATION, id=ID("0_2"))
            label_0_1_1 = LabelEntity(name="label_0_1_1", domain=Domain.CLASSIFICATION, id=ID("0_1_1"))
            non_empty_labels = [label_0, label_0_1, label_0_1_1, label_0_2]
            empty_labels = [
                LabelEntity(
                    name="empty label",
                    domain=Domain.CLASSIFICATION,
                    is_empty=True,
                    id=ID("3"),
                )
            ]
            # Checking "AnnotationSceneEntity" returned by "convert_to_annotation" for
            # "ClassificationToAnnotationConverter" initialized with label group with several non-empty labels
            label_group = LabelGroup(
                name="Classification labels group",
                labels=non_empty_labels + empty_labels,
            )
            label_schema = LabelSchemaEntity(label_groups=[label_group])
            label_schema.add_child(parent=label_0, child=label_0_1)
            label_schema.add_child(parent=label_0, child=label_0_2)
            label_schema.add_child(parent=label_0_1, child=label_0_1_1)
            converter = ClassificationToAnnotationConverter(label_schema=label_schema)
            predictions = [(0, 0.9), (1, 0.8), (2, 0.94), (3, 0.86)]
            predictions_to_annotations = converter.convert_to_annotation(predictions)
            check_annotation_scene(annotation_scene=predictions_to_annotations, expected_length=1)
            check_annotation(
                actual_annotation=predictions_to_annotations.annotations[0],
                expected_labels=[
                    ScoredLabel(label=label_0, probability=0.9),
                    ScoredLabel(label=label_0_1, probability=0.8),
                    ScoredLabel(label=label_0_1_1, probability=0.94),
                    ScoredLabel(label=label_0_2, probability=0.86),
                ],
            )
            # Checking attributes of "AnnotationSceneEntity" returned by "convert_to_annotation" method with
            # "predictions" equal to empty list
            converter = ClassificationToAnnotationConverter(label_schema=label_schema)
            predictions = []
            predictions_to_annotations = converter.convert_to_annotation(predictions)
            check_annotation_scene(annotation_scene=predictions_to_annotations, expected_length=1)
            check_annotation(
                actual_annotation=predictions_to_annotations.annotations[0],
                expected_labels=[ScoredLabel(label=empty_labels[0], probability=1.0)],
            )
            # Checking attributes of "AnnotationSceneEntity" returned by "convert_to_annotation" for
            # "ClassificationToAnnotationConverter" initialized with several LabelGroups
            label_group = LabelGroup(name="Classification labels group", labels=[label_0_1_1])
            other_label_group = LabelGroup(
                name="Other Classification labels group",
                labels=[label_0_1, label_0_2],
            )
            label_schema = LabelSchemaEntity(label_groups=[label_group, other_label_group])

            label_schema.add_child(parent=label_0_1, child=label_0_1_1)
            converter = ClassificationToAnnotationConverter(label_schema=label_schema)
            predictions = [(2, 0.9), (1, 0.8)]
            predictions_to_annotations = converter.convert_to_annotation(predictions)
            check_annotation_scene(annotation_scene=predictions_to_annotations, expected_length=1)
            check_annotation(
                predictions_to_annotations.annotations[0],
                expected_labels=[ScoredLabel(label=label_0_2, probability=0.9)],
            )

    @pytest.mark.components(OtxSdkComponent.OTX_API)
    class TestAnomalyClassificationToAnnotation:
        @pytest.mark.priority_medium
        @pytest.mark.unit
        @pytest.mark.reqids(Requirements.REQ_1)
        def test_anomaly_classification_to_annotation_init(
            self,
        ):
            """
            <b>Description:</b>
            Check "AnomalyClassificationToAnnotationConverter" class initialization

            <b>Input data:</b>
            "AnomalyClassificationToAnnotationConverter" class object with specified "label_schema" parameter

            <b>Expected results:</b>
            Test passes if attributes of initialized "AnomalyClassificationToAnnotationConverter" object are equal
            to expected

            <b>Steps</b>
            1. Check attributes of "AnomalyClassificationToAnnotationConverter" object initialized with non-empty
            labels list
            2. Check attributes of "AnomalyClassificationToAnnotationConverter" object initialized with non-empty
            and empty labels list
            """
            # Checking attributes of "AnomalyClassificationToAnnotationConverter" initialized with non-empty labels
            # list
            non_empty_labels = [
                LabelEntity(name="Normal", domain=Domain.CLASSIFICATION, id=ID("1")),
                LabelEntity(name="Normal", domain=Domain.CLASSIFICATION, id=ID("2")),
                LabelEntity(
                    name="Anomalous",
                    domain=Domain.CLASSIFICATION,
                    id=ID("1"),
                    is_anomalous=True,
                ),
                LabelEntity(
                    name="Anomalous",
                    domain=Domain.CLASSIFICATION,
                    id=ID("2"),
                    is_anomalous=True,
                ),
            ]
            label_group = LabelGroup(name="Classification labels group", labels=non_empty_labels)
            label_schema = LabelSchemaEntity(label_groups=[label_group])
            converter = AnomalyClassificationToAnnotationConverter(label_schema=label_schema)
            assert converter.normal_label == non_empty_labels[0]
            assert converter.anomalous_label == non_empty_labels[2]
            # Checking attributes of "AnomalyClassificationToAnnotationConverter" initialized with non-empty and
            # empty labels list
            empty_labels = [
                LabelEntity(
                    name="Normal",
                    domain=Domain.CLASSIFICATION,
                    is_empty=True,
                    id=ID("3"),
                ),
                LabelEntity(
                    name="Normal",
                    domain=Domain.CLASSIFICATION,
                    is_empty=True,
                    id=ID("4"),
                ),
                LabelEntity(
                    name="Anomalous",
                    domain=Domain.CLASSIFICATION,
                    is_empty=True,
                    id=ID("3"),
                ),
                LabelEntity(
                    name="Anomalous",
                    domain=Domain.CLASSIFICATION,
                    is_empty=True,
                    id=ID("4"),
                ),
            ]
            label_group = LabelGroup(
                name="Anomaly classification labels group",
                labels=non_empty_labels + empty_labels,
            )
            label_schema = LabelSchemaEntity(label_groups=[label_group])
            converter = AnomalyClassificationToAnnotationConverter(label_schema=label_schema)
            assert converter.normal_label == non_empty_labels[0]
            assert converter.anomalous_label == non_empty_labels[2]

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_anomaly_classification_to_annotation_convert(
        self,
    ):
        """
        <b>Description:</b>
        Check "AnomalyClassificationToAnnotationConverter" class "convert_to_annotation" method

        <b>Input data:</b>
        "AnomalyClassificationToAnnotationConverter" class object, "predictions" array

        <b>Expected results:</b>
        Test passes if "AnnotationSceneEntity" object returned by "convert_to_annotation" method is equal to
        expected

        <b>Steps</b>
        1. Check attributes of "AnnotationSceneEntity" object returned by "convert_to_annotation" method for
        "metadata" dictionary with specified "threshold" key
        2. Check attributes of "AnnotationSceneEntity" object returned by "convert_to_annotation" method for
        "metadata" dictionary without specified "threshold" key
        """

        def check_annotation(actual_annotation: Annotation, expected_labels: list):
            assert isinstance(actual_annotation, Annotation)
            assert actual_annotation.get_labels() == expected_labels
            assert isinstance(actual_annotation.shape, Rectangle)
            assert Rectangle.is_full_box(rectangle=actual_annotation.shape)

        non_empty_labels = [
            LabelEntity(name="Normal", domain=Domain.CLASSIFICATION, id=ID("1")),
            LabelEntity(
                name="Anomalous",
                domain=Domain.CLASSIFICATION,
                id=ID("2"),
                is_anomalous=True,
            ),
        ]
        label_group = LabelGroup(name="Anomaly classification labels group", labels=non_empty_labels)
        label_schema = LabelSchemaEntity(label_groups=[label_group])
        converter = AnomalyClassificationToAnnotationConverter(label_schema=label_schema)
        predictions = np.array([0.7])
        # Checking attributes of "AnnotationSceneEntity" returned by "convert_to_annotation" for "metadata" with
        # specified "threshold" key
        metadata = {
            "non-required key": 1,
            "other non-required key": 2,
            "threshold": 0.8,
        }
        predictions_to_annotations = converter.convert_to_annotation(predictions=predictions, metadata=metadata)
        check_annotation_scene(annotation_scene=predictions_to_annotations, expected_length=1)
        check_annotation(
            predictions_to_annotations.annotations[0],
            expected_labels=[ScoredLabel(label=non_empty_labels[0], probability=0.7)],
        )
        # Checking attributes of "AnnotationSceneEntity" returned by "convert_to_annotation" for "metadata" without
        # specified "threshold" key
        metadata = {"non-required key": 1, "other non-required key": 2}
        predictions_to_annotations = converter.convert_to_annotation(predictions=predictions, metadata=metadata)
        check_annotation_scene(annotation_scene=predictions_to_annotations, expected_length=1)
        check_annotation(
            predictions_to_annotations.annotations[0],
            expected_labels=[ScoredLabel(label=non_empty_labels[1], probability=0.7)],
        )



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest

from otx.algorithms.classification.adapters.mmcls.models.necks import MMOVNeck
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.core.ov.models.mmcls.test_helpers import create_ov_model


class TestMMOVNeck:
    @pytest.fixture(autouse=True)
    def setup(self):
        ov_model = create_ov_model()
        self.model = MMOVNeck(
            model_path_or_model=ov_model,
        )

    @e2e_pytest_unit
    def test_parser(self):
        pass



"""Init file for otx.algorithms.common.adapters.mmdeploy.utils."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .mmdeploy import is_mmdeploy_enabled
from .onnx import prepare_onnx_for_openvino, remove_nodes_by_op_type
from .utils import numpy_2_list, sync_batchnorm_2_batchnorm

__all__ = [
    "is_mmdeploy_enabled",
    "sync_batchnorm_2_batchnorm",
    "numpy_2_list",
    "prepare_onnx_for_openvino",
    "remove_nodes_by_op_type",
]



"""This module define the Explain entity."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from dataclasses import dataclass
from typing import Any, Callable, Optional


# pylint: disable=unused-argument
def default_progress_callback(progress: int, score: Optional[float] = None):
    """This is the default progress callback for OptimizationParameters."""


@dataclass
class ExplainParameters:
    """Explain parameters.

    Attributes:
        explainer: Explain algorithm to be used in explanation mode.
            Will be converted automatically to lowercase.
        process_saliency_maps: Processing of saliency map includes (1) resize to input image resolution
            and (2) apply a colormap.
        explain_predicted_classes: Provides explanations only for predicted classes.
            Otherwise, explain all classes.
    """

    update_progress: Callable[[int, Optional[float]], Any] = default_progress_callback

    explainer: str = ""
    process_saliency_maps: bool = False
    explain_predicted_classes: bool = True



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-Lggggcense-Identifier: Apache-2.0
#

import numpy as np
import openvino.runtime as ov
import pytest
import torch

from otx.core.ov.ops.infrastructures import ConstantV0, ParameterV0, ResultV0
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestParameterV0:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 3, 256, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_invalid_attr(self):
        with pytest.raises(ValueError):
            ParameterV0("dummy", shape=(self.shape,), element_type="error")

    @e2e_pytest_unit
    def test_forward(self):
        op = ParameterV0("dummy", shape=(self.shape,))
        assert torch.equal(self.input, op(self.input))

        op = ParameterV0("dummy", shape=(self.shape,), permute=(0, 2, 3, 1))
        assert torch.equal(self.input.permute(0, 2, 3, 1), op(self.input))

        op = ParameterV0("dummy", shape=(self.shape,), verify_shape=True)
        assert torch.equal(self.input, op(self.input))

        op = ParameterV0("dummy", shape=((-1, 3, -1, -1),), verify_shape=True)
        assert torch.equal(self.input, op(self.input))

        with pytest.raises(AssertionError):
            op(self.input.permute(0, 2, 3, 1))

    @e2e_pytest_unit
    def test_from_ov(self):
        op_ov = ov.opset10.parameter([-1, 256, 256, 3], ov.Type.f32)
        op_ov.set_layout(ov.Layout("NHWC"))
        op = ParameterV0.from_ov(op_ov)

        assert isinstance(op, ParameterV0)
        assert op.attrs.element_type == "f32"
        assert op.attrs.layout == ("N", "C", "H", "W")
        assert op.attrs.permute == (0, 2, 3, 1)
        assert op.attrs.shape == ((-1, 3, 256, 256),)


class TestResultV0:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (4, 3, 256, 256)
        self.input = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_forward(self):
        op = ResultV0("dummy", shape=(self.shape,))
        assert torch.equal(self.input, op(self.input))


class TestConstantV0:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.shape = (512, 256)
        self.data = torch.randn(self.shape)

    @e2e_pytest_unit
    def test_invalid_attr(self):
        with pytest.raises(KeyError):
            ConstantV0("dummy", shape=(self.shape,))

    @e2e_pytest_unit
    def test_forward(self):
        op = ConstantV0("dummy", data=self.data, shape=(self.shape,), is_parameter=False)
        assert isinstance(op.data, torch.Tensor)
        assert torch.equal(self.data, op())

        op = ConstantV0("dummy", data=self.data, shape=(self.shape,), is_parameter=True)
        assert isinstance(op.data, torch.nn.parameter.Parameter)
        assert torch.equal(self.data, op())

    @e2e_pytest_unit
    def test_from_ov(self):
        op_ov = ov.opset10.constant(self.data.numpy().astype(np.uint64), ov.Type.u64)
        op = ConstantV0.from_ov(op_ov)
        assert isinstance(op, ConstantV0)
        assert op.attrs.shape == self.shape
        assert not op.attrs.is_parameter
        assert isinstance(op.data, torch.Tensor)

        op_ov = ov.opset10.constant(self.data.numpy(), ov.Type.f32)
        op = ConstantV0.from_ov(op_ov)
        assert isinstance(op, ConstantV0)
        assert op.attrs.shape == self.shape
        assert not op.attrs.is_parameter
        assert isinstance(op.data, torch.Tensor)

        op_ov = ov.opset10.constant(self.data.numpy(), ov.Type.f32)
        data = ov.opset10.parameter([4, 512], ov.Type.f32)
        ov.opset10.matmul(data, op_ov, False, False)
        op = ConstantV0.from_ov(op_ov)
        assert isinstance(op, ConstantV0)
        assert op.attrs.shape == self.shape
        assert op.attrs.is_parameter
        assert isinstance(op.data, torch.nn.parameter.Parameter)



"""Module for defining ext loader."""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import importlib


def load_ext(path, funcs):
    """A function that loads module and verifies that the specified functions are present in the module.

    :param path: str, the file path of the module to load.
    :param funcs: list of str, the names of the functions to verify in the loaded module.
    :return: the loaded module object.
    :raises: AssertionError if any of the specified functions are missing from the loaded module.
    """
    ext = importlib.import_module(path)
    for fun in funcs:
        assert hasattr(ext, fun), f"{fun} miss in module {path}"

    return ext



"""Module of otx.core.ov.ops."""
# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from .activations import (
    ClampV0,
    EluV0,
    ExpV0,
    GeluV7,
    HardSigmoidV0,
    HSigmoidV5,
    HSwishV4,
    MishV4,
    PReluV0,
    ReluV0,
    SeluV0,
    SigmoidV0,
    SoftMaxV0,
    SoftMaxV1,
    SwishV4,
    TanhV0,
)
from .arithmetics import AddV1, DivideV1, MultiplyV1, SubtractV1, TanV0
from .builder import OPS, OperationRegistry
from .convolutions import ConvolutionV1, GroupConvolutionV1
from .generation import RangeV4
from .image_processings import InterpolateV4
from .infrastructures import ConstantV0, ParameterV0, ResultV0
from .matmuls import EinsumV7, MatMulV0
from .movements import (
    BroadcastV3,
    ConcatV0,
    GatherV0,
    GatherV1,
    PadV1,
    ScatterNDUpdateV3,
    ScatterUpdateV3,
    ShuffleChannelsV0,
    SplitV1,
    StridedSliceV1,
    TileV0,
    TransposeV1,
    VariadicSplitV1,
)
from .normalizations import (
    MVNV6,
    BatchNormalizationV0,
    LocalResponseNormalizationV0,
    NormalizeL2V0,
)
from .object_detections import (
    DetectionOutputV0,
    PriorBoxClusteredV0,
    PriorBoxV0,
    ProposalV4,
    RegionYoloV0,
    ROIPoolingV0,
)
from .op import Attribute, Operation
from .poolings import AvgPoolV1, MaxPoolV0
from .reductions import ReduceMeanV1, ReduceMinV1, ReduceProdV1, ReduceSumV1
from .shape_manipulations import ReshapeV1, ShapeOfV0, ShapeOfV3, SqueezeV0, UnsqueezeV0
from .sorting_maximization import NonMaxSuppressionV5, NonMaxSuppressionV9, TopKV3
from .type_conversions import ConvertV0

__all__ = [
    "SoftMaxV0",
    "SoftMaxV1",
    "ReluV0",
    "SwishV4",
    "SigmoidV0",
    "ClampV0",
    "PReluV0",
    "TanhV0",
    "EluV0",
    "SeluV0",
    "MishV4",
    "HSwishV4",
    "HSigmoidV5",
    "ExpV0",
    "HardSigmoidV0",
    "GeluV7",
    "MultiplyV1",
    "DivideV1",
    "AddV1",
    "SubtractV1",
    "TanV0",
    "OPS",
    "OperationRegistry",
    "ConvolutionV1",
    "GroupConvolutionV1",
    "RangeV4",
    "InterpolateV4",
    "ParameterV0",
    "ResultV0",
    "ConstantV0",
    "MatMulV0",
    "EinsumV7",
    "PadV1",
    "ConcatV0",
    "TransposeV1",
    "GatherV0",
    "GatherV1",
    "StridedSliceV1",
    "SplitV1",
    "VariadicSplitV1",
    "ShuffleChannelsV0",
    "BroadcastV3",
    "ScatterNDUpdateV3",
    "ScatterUpdateV3",
    "TileV0",
    "BatchNormalizationV0",
    "LocalResponseNormalizationV0",
    "NormalizeL2V0",
    "MVNV6",
    "ProposalV4",
    "ROIPoolingV0",
    "DetectionOutputV0",
    "RegionYoloV0",
    "PriorBoxV0",
    "PriorBoxClusteredV0",
    "Operation",
    "Attribute",
    "MaxPoolV0",
    "AvgPoolV1",
    "ReduceMeanV1",
    "ReduceProdV1",
    "ReduceMinV1",
    "ReduceSumV1",
    "SqueezeV0",
    "UnsqueezeV0",
    "ReshapeV1",
    "ShapeOfV0",
    "ShapeOfV3",
    "TopKV3",
    "NonMaxSuppressionV5",
    "NonMaxSuppressionV9",
    "ConvertV0",
]



"""Initial file for mmdetection hooks."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .load_pipelines import LoadAnnotationFromOTXDataset, LoadImageFromOTXDataset
from .torchvision2mmdet import (
    BranchImage,
    ColorJitter,
    NDArrayToPILImage,
    NDArrayToTensor,
    PILImageToNDArray,
    RandomApply,
    RandomErasing,
    RandomGaussianBlur,
    RandomGrayscale,
)

__all__ = [
    "LoadImageFromOTXDataset",
    "LoadAnnotationFromOTXDataset",
    "ColorJitter",
    "RandomGrayscale",
    "RandomErasing",
    "RandomGaussianBlur",
    "RandomApply",
    "NDArrayToTensor",
    "NDArrayToPILImage",
    "PILImageToNDArray",
    "BranchImage",
]



"""Tests for MPA Class-Incremental Learning for object detection with OTX CLI"""
# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import copy
import os

import pytest

from otx.api.entities.model_template import parse_model_template
from otx.cli.registry import Registry
from tests.test_suite.e2e_test_system import e2e_pytest_component
from tests.test_suite.run_test_command import (
    get_template_dir,
    nncf_eval_openvino_testing,
    nncf_eval_testing,
    nncf_export_testing,
    nncf_optimize_testing,
    nncf_validate_fq_testing,
    otx_demo_deployment_testing,
    otx_demo_openvino_testing,
    otx_demo_testing,
    otx_deploy_openvino_testing,
    otx_eval_deployment_testing,
    otx_eval_openvino_testing,
    otx_eval_testing,
    otx_explain_openvino_testing,
    otx_explain_testing,
    otx_export_testing,
    otx_hpo_testing,
    otx_resume_testing,
    otx_train_testing,
    pot_eval_testing,
    pot_optimize_testing,
    pot_validate_fq_testing,
)

args = {
    "--train-data-roots": "tests/assets/small_objects",
    "--val-data-roots": "tests/assets/small_objects",
    "--test-data-roots": "tests/assets/small_objects",
    "--input": "tests/assets/small_objects/images/train",
    "train_params": [
        "params",
        "--learning_parameters.num_iters",
        "10",
        "--learning_parameters.batch_size",
        "4",
        "--tiling_parameters.enable_tiling",
        "1",
        "--tiling_parameters.enable_adaptive_params",
        "1",
    ],
}

# Training params for resume, num_iters*2
resume_params = [
    "params",
    "--learning_parameters.num_iters",
    "15",
    "--learning_parameters.batch_size",
    "4",
]

otx_dir = os.getcwd()

TT_STABILITY_TESTS = os.environ.get("TT_STABILITY_TESTS", False)
if TT_STABILITY_TESTS:
    default_template = parse_model_template(
        os.path.join("otx/algorithms/detection/configs", "detection", "mobilenetv2_atss", "template.yaml")
    )
    templates = [default_template] * 100
    templates_ids = [template.model_template_id + f"-{i+1}" for i, template in enumerate(templates)]
else:
    templates = Registry("otx/algorithms/detection").filter(task_type="DETECTION").templates
    templates_ids = [template.model_template_id for template in templates]


class TestToolsTilingDetection:
    @e2e_pytest_component
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_train(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        otx_train_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_resume(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det/test_resume"
        otx_resume_testing(template, tmp_dir_path, otx_dir, args)
        template_work_dir = get_template_dir(template, tmp_dir_path)
        args1 = copy.deepcopy(args)
        args1["train_params"] = resume_params
        args1[
            "--resume-from"
        ] = f"{template_work_dir}/trained_for_resume_{template.model_template_id}/models/weights.pth"
        otx_resume_testing(template, tmp_dir_path, otx_dir, args1)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    @pytest.mark.parametrize("dump_features", [True, False])
    def test_otx_export(self, template, tmp_dir_path, dump_features):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        otx_export_testing(template, tmp_dir_path, dump_features)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_export_fp16(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        otx_export_testing(template, tmp_dir_path, half_precision=True)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_eval(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        otx_eval_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    @pytest.mark.parametrize("half_precision", [True, False])
    def test_otx_eval_openvino(self, template, tmp_dir_path, half_precision):
        if template.name in ["SSD", "ATSS"]:
            pytest.skip(reason="[CVS-108291] Tiling ATSS, SSD show performance drop")
        tmp_dir_path = tmp_dir_path / "tiling_det"
        otx_eval_openvino_testing(template, tmp_dir_path, otx_dir, args, threshold=0.05, half_precision=half_precision)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_explain(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        otx_explain_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_explain_openvino(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        otx_explain_openvino_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_demo(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        otx_demo_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_demo_openvino(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        otx_demo_openvino_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_deploy_openvino(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        otx_deploy_openvino_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_eval_deployment(self, template, tmp_dir_path):
        if template.name in ["ATSS", "SSD"]:
            pytest.skip(reason="[CVS-108291] Tiling ATSS, SSD show performance drop")
        tmp_dir_path = tmp_dir_path / "tiling_det"
        otx_eval_deployment_testing(template, tmp_dir_path, otx_dir, args, threshold=0.0)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_otx_demo_deployment(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        otx_demo_deployment_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    @pytest.mark.skip("[CVS-108810] Tiling w/ HPO fails in CI")
    def test_otx_hpo(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det/test_hpo"
        otx_hpo_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_nncf_optimize(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        if template.entrypoints.nncf is None:
            pytest.skip("nncf entrypoint is none")

        nncf_optimize_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_nncf_export(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        if template.entrypoints.nncf is None:
            pytest.skip("nncf entrypoint is none")

        nncf_export_testing(template, tmp_dir_path)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_nncf_validate_fq(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        if template.entrypoints.nncf is None:
            pytest.skip("nncf entrypoint is none")

        nncf_validate_fq_testing(template, tmp_dir_path, otx_dir, "detection", type(self).__name__)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_nncf_eval(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        if template.entrypoints.nncf is None:
            pytest.skip("nncf entrypoint is none")

        nncf_eval_testing(template, tmp_dir_path, otx_dir, args, threshold=0.01)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_nncf_eval_openvino(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        if template.entrypoints.nncf is None:
            pytest.skip("nncf entrypoint is none")

        nncf_eval_openvino_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_pot_optimize(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        pot_optimize_testing(template, tmp_dir_path, otx_dir, args)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_pot_validate_fq(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        pot_validate_fq_testing(template, tmp_dir_path, otx_dir, "detection", type(self).__name__)

    @e2e_pytest_component
    @pytest.mark.skipif(TT_STABILITY_TESTS, reason="This is TT_STABILITY_TESTS")
    @pytest.mark.parametrize("template", templates, ids=templates_ids)
    def test_pot_eval(self, template, tmp_dir_path):
        tmp_dir_path = tmp_dir_path / "tiling_det"
        pot_eval_testing(template, tmp_dir_path, otx_dir, args)



import argparse
import json
import os
from typing import Dict, Union

import pandas as pd

from tests.regression.regression_test_helpers import (
    ANOMALY_DATASET_CATEGORIES,
    LABEL_TYPES,
    TRAIN_TYPES,
)

ANOMALY_DATA = {
    "Task type": [],
    "MVTec Category": [],
    "Model": [],
    "train": [],
    "export": [],
    "deploy": [],
    "nncf": [],
    "pot": [],
    "train E2E Time (Sec.)": [],
    "export E2E Time (Sec.)": [],
    "deploy E2E Time (Sec.)": [],
    "nncf E2E Time (Sec.)": [],
    "pot E2E Time (Sec.)": [],
    "train Eval Time (Sec.)": [],
    "export Eval Time (Sec.)": [],
    "deploy Eval Time (Sec.)": [],
    "nncf Eval Time (Sec.)": [],
    "pot Eval Time (Sec.)": [],
}

NON_ANOMALY_DATA = {
    "Task type": [],
    "Train type": [],
    "Label type": [],
    "Model": [],
    "train": [],
    "export": [],
    "deploy": [],
    "nncf": [],
    "pot": [],
    "train E2E Time (Sec.)": [],
    "export E2E Time (Sec.)": [],
    "deploy E2E Time (Sec.)": [],
    "nncf E2E Time (Sec.)": [],
    "pot E2E Time (Sec.)": [],
    "train Eval Time (Sec.)": [],
    "export Eval Time (Sec.)": [],
    "deploy Eval Time (Sec.)": [],
    "nncf Eval Time (Sec.)": [],
    "pot Eval Time (Sec.)": [],
}


def get_metric_dict(dict_type, idx: int, model: str):
    if dict_type:
        return dict_type[idx][model]
    else:
        return "-"


def get_metric_items(input_data):
    if isinstance(input_data, dict):
        return_list = []
        for k, v in input_data.items():
            return_list.append((k, v))
        return return_list
    else:
        return "-"


def filter_task(root: str) -> Dict[str, str]:
    """Find prpoer task and task_key."""
    task = root.split("/")[-1]
    if "tiling" in task:
        task_key = "_".join(task.split("_")[1:])
    else:
        task_key = task
    return {"task_key": task_key, "task": task}


def is_anomaly_task(task: str) -> bool:
    """Returns True if task is anomaly."""
    return "anomaly" in task


def fill_model_performance(items: Union[list, str], test_type: str, result_data: dict):
    if isinstance(items, list):
        result_data[test_type].append(f"{items[0][0]}: {items[0][1]}")
        if test_type == "train":
            result_data[f"{test_type} E2E Time (Sec.)"].append(f"{items[2][1]}")
            result_data[f"{test_type} Eval Time (Sec.)"].append(f"{items[3][1]}")
        else:
            result_data[f"{test_type} E2E Time (Sec.)"].append(f"{items[1][1]}")
            result_data[f"{test_type} Eval Time (Sec.)"].append(f"{items[2][1]}")
    else:
        result_data[test_type].append(items)
        result_data[f"{test_type} E2E Time (Sec.)"].append(items)
        result_data[f"{test_type} Eval Time (Sec.)"].append(items)


def summarize_non_anomaly_data(task: str, task_key: str, json_data: dict, result_data: dict) -> dict:
    """Make DataFrame by gathering all results."""
    for label_type in LABEL_TYPES:
        for train_type in TRAIN_TYPES:
            task_data = json_data[task_key][label_type][train_type]

            train_data = task_data.get("train")
            if train_data is None:
                raise ValueError("Train data can't be empty.")
            export_data = task_data.get("export")
            deploy_data = task_data.get("deploy")
            nncf_data = task_data.get("nncf")
            pot_data = task_data.get("pot")

            for i, per_model_data in enumerate(train_data):
                for model in per_model_data:
                    train_items = get_metric_items(get_metric_dict(train_data, i, model))
                    export_items = get_metric_items(get_metric_dict(export_data, i, model))
                    deploy_items = get_metric_items(get_metric_dict(deploy_data, i, model))
                    nncf_items = get_metric_items(get_metric_dict(nncf_data, i, model))
                    pot_items = get_metric_items(get_metric_dict(pot_data, i, model))

                    result_data["Task type"].append(task)
                    result_data["Train type"].append(train_type)
                    result_data["Label type"].append(label_type)
                    result_data["Model"].append(model)

                    fill_model_performance(train_items, "train", result_data)
                    fill_model_performance(export_items, "export", result_data)
                    fill_model_performance(deploy_items, "deploy", result_data)
                    fill_model_performance(nncf_items, "nncf", result_data)
                    fill_model_performance(pot_items, "pot", result_data)


def summarize_anomaly_data(task: str, task_key: str, json_data: dict, result_data: dict) -> dict:
    """Make DataFrame by gathering all results."""
    task_data = json_data[task_key]

    train_data = task_data.get("train")
    if train_data is None:
        raise ValueError("Train data can't be empty.")
    export_data = task_data.get("export")
    deploy_data = task_data.get("deploy")
    nncf_data = task_data.get("nncf")
    pot_data = task_data.get("pot")

    for anomaly_category in ANOMALY_DATASET_CATEGORIES:
        train_cat_data = train_data.get(anomaly_category)
        export_cat_data = export_data.get(anomaly_category)
        deploy_cat_data = deploy_data.get(anomaly_category)
        nncf_cat_data = nncf_data.get(anomaly_category)
        pot_cat_data = pot_data.get(anomaly_category)

        for i, per_model_data in enumerate(train_cat_data):
            for model in per_model_data:
                train_items = get_metric_items(get_metric_dict(train_cat_data, i, model))
                export_items = get_metric_items(get_metric_dict(export_cat_data, i, model))
                deploy_items = get_metric_items(get_metric_dict(deploy_cat_data, i, model))
                nncf_items = get_metric_items(get_metric_dict(nncf_cat_data, i, model))
                pot_items = get_metric_items(get_metric_dict(pot_cat_data, i, model))

                result_data["Task type"].append(task)
                result_data["MVTec Category"].append(anomaly_category)
                result_data["Model"].append(model)

                fill_model_performance(train_items, "train", result_data)
                fill_model_performance(export_items, "export", result_data)
                fill_model_performance(deploy_items, "deploy", result_data)
                fill_model_performance(nncf_items, "nncf", result_data)
                fill_model_performance(pot_items, "pot", result_data)


def save_file(result_data: dict, output_path: str):
    df = pd.DataFrame(result_data)
    df.to_csv(output_path)


def parse_args():
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("--input_path", default="/tmp/regression_test_results", type=str)
    parser.add_argument("--output_path", default="/tmp", type=str)
    return parser.parse_args()


def summarize_data(args):
    input_path = args.input_path

    for root, _, files in os.walk(input_path):
        for result_file in files:
            task_dict = filter_task(root)
            task_key, task = task_dict["task_key"], task_dict["task"]

            json_file_path = os.path.join(root, result_file)
            with open(json_file_path, "r") as f:
                json_data = json.load(f)

            if is_anomaly_task(task) is True:
                summarize_anomaly_data(task, task_key, json_data, ANOMALY_DATA)
            else:
                summarize_non_anomaly_data(task, task_key, json_data, NON_ANOMALY_DATA)

    save_file(ANOMALY_DATA, f"{args.output_path}/anomaly_results.csv")
    save_file(NON_ANOMALY_DATA, f"{args.output_path}/non_anomaly_results.csv")


if __name__ == "__main__":
    args = parse_args()
    summarize_data(args)



import pytest
import torch

from otx.algorithms.segmentation.adapters.mmseg.models.losses.cross_entropy_loss_with_ignore import (
    CrossEntropyLossWithIgnore,
)
from mmseg.models.losses import CrossEntropyLoss

from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestCrossEntropyLosWithIgnore:
    @pytest.fixture(autouse=True)
    def setup(self):
        self.mock_score = torch.rand([1, 2, 5, 5])
        self.mock_gt = torch.zeros((1, 5, 5), dtype=torch.long)
        self.mock_gt[::2, 1::2, :] = 1
        self.mock_gt[1::2, ::2, :] = 1

        self.loss_f = CrossEntropyLossWithIgnore()

    @e2e_pytest_unit
    def test_is_label_ignored(self):
        loss = self.loss_f(self.mock_score, self.mock_gt, reduction_override="none")
        assert type(loss) == torch.Tensor

        mock_valid_label_mask = torch.Tensor([[1, 0]])
        loss_ignore = self.loss_f(
            self.mock_score, self.mock_gt, reduction_override="none", valid_label_mask=mock_valid_label_mask
        )

        assert torch.all(loss_ignore[::2, 1::2, :] == 0)
        assert torch.all(loss_ignore[1::2, ::2, :] == 0)

        assert torch.equal(loss, loss_ignore) is False

    @e2e_pytest_unit
    def test_is_equal_to_ce_loss(self):
        loss_f_mmseg = CrossEntropyLoss()

        loss_1 = loss_f_mmseg(self.mock_score, self.mock_gt)
        loss_2 = self.loss_f(self.mock_score, self.mock_gt)
        loss_3 = self.loss_f(self.mock_score, self.mock_gt, valid_label_mask=torch.Tensor([1, 1]))

        assert loss_1 == loss_2
        assert loss_2 == loss_3



"""OTX Algorithms - Classification Models."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from .classifiers import BYOL, SAMImageClassifier, SemiSLClassifier, SupConClassifier
from .heads import (
    ClsHead,
    ConstrastiveHead,
    ConvClsHead,
    CustomHierarchicalLinearClsHead,
    CustomHierarchicalNonLinearClsHead,
    CustomLinearClsHead,
    CustomMultiLabelLinearClsHead,
    CustomMultiLabelNonLinearClsHead,
    CustomNonLinearClsHead,
    MMOVClsHead,
    SemiLinearMultilabelClsHead,
    SemiNonLinearMultilabelClsHead,
    SupConClsHead,
)
from .losses import (
    AsymmetricAngularLossWithIgnore,
    AsymmetricLossWithIgnore,
    BarlowTwinsLoss,
    CrossEntropyLossWithIgnore,
    IBLoss,
)
from .necks import SelfSLMLP

__all__ = [
    "BYOL",
    "SAMImageClassifier",
    "SemiSLClassifier",
    "SupConClassifier",
    "CustomLinearClsHead",
    "CustomNonLinearClsHead",
    "CustomMultiLabelNonLinearClsHead",
    "CustomMultiLabelLinearClsHead",
    "CustomHierarchicalLinearClsHead",
    "CustomHierarchicalNonLinearClsHead",
    "AsymmetricAngularLossWithIgnore",
    "SemiLinearMultilabelClsHead",
    "SemiNonLinearMultilabelClsHead",
    "MMOVClsHead",
    "ConvClsHead",
    "ClsHead",
    "AsymmetricLossWithIgnore",
    "BarlowTwinsLoss",
    "IBLoss",
    "CrossEntropyLossWithIgnore",
    "SelfSLMLP",
    "ConstrastiveHead",
    "SupConClsHead",
]
